package edu.washington.cs.radishsim;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.NavigableSet;
import java.util.TreeSet;

class Check implements Comparable<Check> {

   final long  start;
   final long  stop;
   final CpuId owner;

   Check(long a, long b, CpuId c) {
      this.start = a;
      this.stop = b;
      this.owner = c;
   }

   /** Compare Checks by their start time. */
   @Override
   public int compareTo(Check c) {
      if ( !(c instanceof Check) ) throw new ClassCastException();
      if ( this.start < c.start ) return -1;
      else if ( this.start == c.start ) return 0;
      else return 1;
   }

   @Override
   public String toString() {
      return "[" + this.start + "," + this.stop + "]@" + this.owner.toString();
   }
}

/**
 * @deprecated This class is no longer used. Async checks should be simulated as a list of actual
 *             memory accesses running on CheckerProcessor cores.
 * @see CheckerProcessor
 */
public class AsyncSwCheckManager {

   private final NavigableSet<Check> checkStartTime = new TreeSet<Check>();

   public AsyncSwCheckManager() {}

   /** flush out checks that have completed */
   private <Line extends RadishLine> void clearCompletedChecks(final int latency, final long now) {
      final Check nowCheck = new Check(now - latency, now - latency + 1, null);

      NavigableSet<Check> completed = checkStartTime.headSet(nowCheck, true);
      completed.clear();
   }

   /**
    * Queue up another async check.
    * 
    * @return the delay (in cycles) of this operation (non-zero when the async check queue is
    *         already full)
    */
   public <Line extends RadishLine> long addAsyncCheck(int latency, final long now, CpuId cpu,
                                                       boolean test) {
      clearCompletedChecks(latency, now);

      final Check incoming = new Check(now, now + latency, cpu);
      final Check before = checkStartTime.floor(incoming); // <=
      final Check after = checkStartTime.ceiling(incoming); // >=
      if ( (before == null || before.stop <= incoming.start)
           && (after == null || after.start >= incoming.stop) ) {
         // we have capacity
         if ( !test ) checkStartTime.add(incoming);
         return 0;

      } else {
         // wait for next check to finish, then enqueue
         Check nextCheck = (before == null) ? after : before;
         List<Check> futureChecks = new LinkedList<Check>(checkStartTime.tailSet(nextCheck, true)); // >=
         assert futureChecks.size() > 0;

         for ( ListIterator<Check> i = futureChecks.listIterator(); i.hasNext(); ) {
            Check a = i.next();

            if ( !i.hasNext() ) {
               // a is the check scheduled farthest in the future; schedule incoming after a
               if ( !test ) checkStartTime.add(new Check(a.stop, a.stop + latency, cpu));
               assert a.stop > now;
               return a.stop - now;
            }

            Check b = i.next();
            // can we fit incoming between a and b?
            if ( a.stop + latency <= b.start ) {
               if ( !test ) checkStartTime.add(new Check(a.stop, a.stop + latency, cpu));
               assert a.stop > now;
               return a.stop - now;
            }

            i.previous();
         }
      }

      throw new IllegalStateException("Impossible!");
   }

   /**
    * Flushes all pending async checks for the given processor.
    * 
    * @return the number of cycles that were required to flush the checks
    */
   public <Line extends RadishLine> long flushAsyncChecks(final int latency, final long now,
                                                          CpuId cpu) {
      clearCompletedChecks(latency, now);

      Iterator<Check> i = checkStartTime.descendingIterator();
      Check mostRecent = null;
      while ( i.hasNext() ) {
         Check c = i.next();
         if ( c.owner.equals(cpu) ) {
            mostRecent = c;
            break;
         }
      }

      // no pending checks for this processor
      if ( mostRecent == null ) return 0;

      // this processor has some pending checks

      // wait for the most recent check to complete
      final long wait = mostRecent.stop - now;

      // all pending checks are finished now
      i = checkStartTime.iterator();
      while ( i.hasNext() ) {
         Check c = i.next();
         if ( c.owner.equals(cpu) ) {
            i.remove();
         }
      }

      assert wait > 0;
      return wait;
   }

}
