package edu.washington.cs.radishsim;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Arrays;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import com.google.common.collect.*;

import edu.washington.cs.radishsim.Machine.LeafReadConfiguration;
import edu.washington.cs.radishsim.Permissions.MergingResult;
import edu.washington.cs.radishsim.RadishLine.ByteMetadata;
import static edu.washington.cs.radishsim.MemorySystemConstants.*;

class SyncObject {
   Epoch[] vc        = null;
   long    timestamp = 0;
}

public class Processor<Line extends RadishLine> implements CacheCallbacks<Line> {

   static final Random                                     rng                    = new Random();

   /** map from sync objects to a vector clock snapshot (shared by all processors) */
   final Map<Long, SyncObject>                             vcOfSyncObject;

   final CpuId                                             id;
   /** Machine reference is shared by all processors */
   final Machine<Line>                                     machine;

   /** L1 caches are always present, and are private to each processor */
   final HierarchicalCache<Line>                           L1cache;
   /** L2 caches are optional, and (if present) are private to each processor */
   final HierarchicalCache<Line>                           L2cache;
   /** the L3 cache is optional, and (if present) is shared by all processors */
   final HierarchicalCache<Line>                           L3cache;

   /**
    * List of all the processors in the system. All processors share the same array object.
    */
   final Processor<Line>[]                                 allProcessors;

   /** List of all checker cores. All processors share the same array object. */
   final CheckerProcessor<Line>[]                          checkerProcs;

   /**
    * Set of byte addresses that are in HW. Addresses not in this set have metadata in software. All
    * caches share the same map.
    */
   final Map<UnownedMetadataByteAddress, InHardwareStatus> inHardwareStatus;

   /** The number of program variables in a given cache line. */
   // final Map<LineAddress, Integer> variablesInLine;

   /**
    * Records the logical time (epoch #) this core perceives from other cores. Initially, all 0's
    * except for a 1 for the entry corresponding to this core. This corresponds to epoch counts
    * beginning at 1.
    */
   final Epoch[]                                           currentVectorClock;

   /** All processors use the same global table */
   final SoftwareMetadata                                  globalMetaDataTable;

   /** Local epochs are represented by base+offset. */
   EpochBase                                               localEpochBase         = EpochBase.INITIAL_EPOCH_BASE;

   final EpochOffsetManager                                epochOffsetManager;

   private final Machine.MachineParams<Line>               params;

   final ProcessorStats                                    stats                  = new ProcessorStats();

   /** The cycle of the previous memory usage check. */
   private long                                            lastCacheUsageSnapshot = 0;
   List<CacheUsageSnapshot>                                cacheUsageSnapshots    = new LinkedList<CacheUsageSnapshot>();

   class ProcessorStats {

      // counters for cache events
      class CacheEventCounter {
         Counter pc_ReadHits;
         Counter pc_ReadMisses;
         Counter pc_WriteHits;
         Counter pc_WriteMisses;

         CacheEventCounter(String prefix) {
            pc_ReadHits = new Counter("pc_" + prefix + "ReadHits");
            pc_ReadMisses = new Counter("pc_" + prefix + "ReadMisses");
            pc_WriteHits = new Counter("pc_" + prefix + "WriteHits");
            pc_WriteMisses = new Counter("pc_" + prefix + "WriteMisses");
         }
      }

      // Counter pc_ReadHits = new Counter("pc_ReadHits");
      Counter           pc_ReadRemoteHits                = new Counter("pc_ReadRemoteHits");
      // Counter pc_ReadMisses = new Counter("pc_ReadMisses");
      // Counter pc_WriteHits = new Counter("pc_WriteHits");
      Counter           pc_WriteRemoteHits               = new Counter("pc_WriteRemoteHits");
      // Counter pc_WriteMisses = new Counter("pc_WriteMisses");
      Counter           pc_UpgradeMisses                 = new Counter("pc_UpgradeMisses");

      CacheEventCounter pc_l1d                           = new CacheEventCounter("Data_L1");
      CacheEventCounter pc_l2d                           = new CacheEventCounter("Data_L2");
      CacheEventCounter pc_l3d                           = new CacheEventCounter("Data_L3");
      CacheEventCounter pc_l1md                          = new CacheEventCounter("Metadata_L1");
      CacheEventCounter pc_l2md                          = new CacheEventCounter("Metadata_L2");
      CacheEventCounter pc_l3md                          = new CacheEventCounter("Metadata_L3");
      Counter           pc_TotalDataReads                = new Counter("pc_TotalDataReads");
      Counter           pc_TotalDataWrites               = new Counter("pc_TotalDataWrites");
      Counter           pc_TotalMemoryAccesses           = new Counter("pc_TotalMemoryAccesses");

      Counter           pc_SyncSources                   = new Counter("pc_SyncSources");
      Counter           pc_SyncTotalSinks                = new Counter("pc_SyncTotalSinks");
      Counter           pc_SyncSourcelessSinks           = new Counter("pc_SyncSourcelessSinks");
      Counter           pc_SyncUnmatchedSinks            = new Counter("pc_SyncUnmatchedSinks");
      Counter           pc_CyclesInMemoryHierarchy       = new Counter("pc_CyclesInMemoryHierarchy");
      Counter           pc_ExecDrivenCycleCount          = new Counter(
                                                                       "pc_ExecutionDrivenCycleCount");

      // counters for Radish events

      Counter           pc_TotalMetadataReads            = new Counter("pc_TotalMetadataReads");
      Counter           pc_TotalMetadataWrites           = new Counter("pc_TotalMetadataWrites");
      Counter           pc_MetadataEvictions             = new Counter("pc_MetadataEvictions");
      Counter           pc_UnimportantMetadataEvictions  = new Counter(
                                                                       "pc_UnimportantMetadataEvictions");
      Counter           pc_LLCDataOverMdEvictions        = new Counter("pc_LLCDataOverMdEvictions");
      Counter           pc_TotallyOrderedMdReads         = new Counter(
                                                                       "pc_TotallyOrderedMetadataReads");
      Counter           pc_EvictedFalseLeafReads         = new Counter("pc_EvictedFalseLeafReads");
      /** only computed when we generate stats */
      Counter           pc_Epochs                        = new Counter("pc_Epochs");
      Counter           pc_EpochRollovers                = new Counter("pc_EpochRollovers");

      /** when our read conflicts with the remote bits */
      Counter           pc_LocalReadViolatesLocalPerms   = new Counter(
                                                                       "pc_LocalReadViolatesPermissions");
      Counter           pc_LocalWriteViolatesLocalPerms  = new Counter(
                                                                       "pc_LocalWriteViolatesPermissions");
      /** when a remote read conflicts with our local bits */
      Counter           pc_RemoteReadViolatesLocalPerms  = new Counter(
                                                                       "pc_RemoteReadConflictsWithLocalBits");
      Counter           pc_RemoteWriteViolatesLocalPerms = new Counter(
                                                                       "pc_RemoteWriteConflictsWithLocalBits");
      /**
       * the number of times we had to consult the hardware VC striped across all the caches, and
       * couldn't piggyback on existing coherence traffic.
       */
      Counter           pc_HwCacheChecks                 = new Counter("pc_HwCacheChecks");
      /** software VC checks triggered */
      Counter           pc_SwVcChecksTotal               = new Counter("pc_SoftwareVcChecksTotal");
      Counter           pc_SwVcChecksTotal_DueToReads    = new Counter(
                                                                       "pc_SoftwareVcChecksTotal_DueToReads");
      Counter           pc_SwVcChecksTotal_DueToWrites   = new Counter(
                                                                       "pc_SoftwareVcChecksTotal_DueToWrites");
      Counter           pc_SwVcChecksAsync_DueToReads    = new Counter(
                                                                       "pc_SoftwareVcChecksAsync_DueToReads");
      Counter           pc_SwVcChecksAsync_DueToWrites   = new Counter(
                                                                       "pc_SoftwareVcChecksAsync_DueToWrites");
      Counter           pc_SwVcChecksSync_DueToReads     = new Counter(
                                                                       "pc_SoftwareVcChecksSync_DueToReads");
      Counter           pc_SwVcChecksSync_DueToWrites    = new Counter(
                                                                       "pc_SoftwareVcChecksSync_DueToWrites");
      /** the total number of is-there-enough-metadata-in-hardware checks performed */
      Counter           pc_HwSufficiencyChecks           = new Counter("pc_HwSufficiencyChecks");

      /** the number of data races we found */
      Counter           pc_DataRaces                     = new Counter("pc_DataRaces");

      Counter           pc_VarsCheckedInSoftwareChecks   = new Counter(
                                                                       "pc_VariablesCheckedInSoftwareChecks");
      Counter           pc_SoftwareChecksWithVarInfo     = new Counter(
                                                                       "pc_SoftwareChecksWithVariableInfo");
      Counter           pc_UselessSwChecks_SwEmpty_Read  = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareEmpty_DueToReads");
      Counter           pc_UselessSwChecks_SwEmpty_Write = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareEmpty_DueToWrites");
      Counter           pc_UselessSwChecks_SwEmpty_Total = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareEmpty_Total");
      Counter           pc_UselessSwChecks_SwStale_Read  = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareStale_DueToReads");
      Counter           pc_UselessSwChecks_SwStale_Write = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareStale_DueToWrites");
      Counter           pc_UselessSwChecks_SwStale_Total = new Counter(
                                                                       "pc_UselessSoftwareChecks_SoftwareStale_Total");
      Counter           pc_CompactMdCompanionChecks      = new Counter(
                                                                       "pc_CompactMdCompanionChecks");
      Counter           pc_CompactMdExpansions           = new Counter("pc_CompactMdExpansions");
   }

   public Processor(Machine.MachineParams<Line> args, Machine<Line> machine, CpuId cpuid,
                    SoftwareMetadata globalTab, Map<Long, SyncObject> syncObjectMap,
                    Processor<Line>[] processors,
                    Map<UnownedMetadataByteAddress, InHardwareStatus> inHwStatus,
                    Map<LineAddress, Integer> varmap, CheckerProcessor<Line>[] checkers) {
      this.params = args;
      this.id = cpuid;
      this.machine = machine;
      this.vcOfSyncObject = syncObjectMap;
      this.allProcessors = processors;
      this.checkerProcs = checkers;
      this.inHardwareStatus = inHwStatus;
      // this.variablesInLine = varmap;
      this.epochOffsetManager = new EpochOffsetManager(args.bitsPerLocalEpochOffset(),
                                                       args.useClockGC());
      this.globalMetaDataTable = globalTab;

      this.currentVectorClock = new Epoch[processors.length];
      for ( int i = 0; i < currentVectorClock.length; i++ ) {
         this.currentVectorClock[i] = Epoch.EMPTY_EPOCH;
      }

      if ( this.id.get() < processors.length ) {
         // don't setup VC entries for the checker procs
         currentVectorClock[this.id.get()] = this.localEpochBase.add(epochOffsetManager.current());
      }

      if ( args.useL3() ) {
         /*
          * NB: hack to get a shared L3. Necessary because we want to have a Processor object handle
          * L3 cache evictions, and we need to supply a CacheCallbacks object to the cache ctor. So
          * all the cache construction has to occur inside the Processor ctor.
          */
         if ( this.allProcessors[0] == null ) {
            assert cpuid.get() == 0;
            // Processor 0 (us) is being constructed, so let's build the shared L3.
            // NB: processor 0 handles L3 evictions!
            this.L3cache = new HierarchicalCache<Line>(args.l3config(), this, null,
                                                       args.lineFactory());
         } else {
            // reuse processor 0's L3 reference
            this.L3cache = this.allProcessors[0].L3cache;
         }
      } else {
         this.L3cache = null;
      }
      if ( args.useL2() ) {
         this.L2cache = new HierarchicalCache<Line>(args.l2config(), this, this.L3cache,
                                                    args.lineFactory());
         this.L1cache = new HierarchicalCache<Line>(args.l1config(), this, this.L2cache,
                                                    args.lineFactory());
      } else {
         this.L2cache = null;
         this.L1cache = new HierarchicalCache<Line>(args.l1config(), this, this.L3cache,
                                                    args.lineFactory());
      }
   }

   @Override
   public String toString() {
      return id.toString();
   }

   /** Update counters that are lazily-computed. */
   public void finalizeCounters() {
      stats.pc_Epochs.set(localEpochBase.add(epochOffsetManager.current()).get());
   }

   public void insnsExecuted(byte n) {
      stats.pc_ExecDrivenCycleCount.incr(n);
   }

   private void memoryCyclesElapsed(int n, DataMemoryAccessResult mor) {
      stats.pc_ExecDrivenCycleCount.incr(n);
      this.stats.pc_CyclesInMemoryHierarchy.incr(n);
      if ( mor != null ) {
         mor.latency += n;
      }
   }

   /** Returns which level is the last-level cache in the system. */
   private CacheLevel llc() {
      if ( L3cache != null ) {
         return CacheLevel.L3;
      } else if ( L2cache != null ) {
         return CacheLevel.L2;
      } else return CacheLevel.L1;
   }

   @Override
   public Line eviction(final Line incoming, final Deque<Line> set, final CacheLevel level) {
      Line toEvict = null;

      if ( params.metadataLLCEvictionPolicy() && level == llc() ) {
         for ( Iterator<Line> i = set.descendingIterator(); i.hasNext(); ) {
            Line candidate = i.next();
            if ( !candidate.isMetadata() ) {
               toEvict = candidate;
               if ( candidate != set.getLast() ) {
                  stats.pc_LLCDataOverMdEvictions.incr();
               }
               break;
            }
         }
         // if we couldn't find any data to evict, evict LRU metadata
         if ( toEvict == null ) toEvict = set.getLast();

      } else if ( params.evictUselessMetadata() ) {
         for ( Iterator<Line> i = set.descendingIterator(); i.hasNext(); ) {
            Line candidate = i.next();
            if ( candidate.isMetadata() && !candidate.evictionWouldDowngradeInHardwareStatus() ) {
               toEvict = candidate;
               if ( candidate != set.getLast() ) {
                  stats.pc_UnimportantMetadataEvictions.incr();
               }
               break;
            }
         }
         if ( toEvict == null ) toEvict = set.getLast();

      } else { // straight LRU
         toEvict = set.getLast();
      }

      boolean evictionFromLLC = (level == llc());
      if ( evictionFromLLC && toEvict.isMetadata() ) {
         evictMetadataToMemory(toEvict, evictionFromLLC, incoming);
      }
      return toEvict;
   }

   public static class DataMemoryAccessResult {

      /** The latency of this memory operation, in cycles. */
      int     latency                    = 0;

      /**
       * true if remote communication happened, due to remote hit or LLC miss, false on a purely
       * local hit
       */
      boolean remoteCommunicatedHappened = false;

      /** Aggregate the result of another memory op into the current result. */
      void aggregate(DataMemoryAccessResult dmar) {
         this.remoteCommunicatedHappened |= dmar.remoteCommunicatedHappened;
         this.latency = Math.max(this.latency, dmar.latency);
      }
   }

   /**
    * Perform a data read specified by the given access.
    * 
    */
   public DataMemoryAccessResult read(final DataAccess access) {

      stats.pc_TotalDataReads.incr();
      stats.pc_TotalMemoryAccesses.incr();

      DataMemoryAccessResult dmaResult = new DataMemoryAccessResult();
      dmaResult.remoteCommunicatedHappened = false;

      if ( RadishSim.XASSERTS ) Verify.cacheExclusivity();

      MemoryResponse<Line> resp = L1cache.request(access.addr());
      Line line = resp.lineHit;

      switch ( resp.whereHit ) {
      case L1:
         stats.pc_l1d.pc_ReadHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.L1_HIT_LATENCY, dmaResult);
         return dmaResult;
      case L2:
         stats.pc_l1d.pc_ReadMisses.incr();
         stats.pc_l2d.pc_ReadHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.L2_HIT_LATENCY, dmaResult);
         return dmaResult;
      case L3:
         stats.pc_l1d.pc_ReadMisses.incr();
         stats.pc_l2d.pc_ReadMisses.incr();
         stats.pc_l3d.pc_ReadHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.L3_HIT_LATENCY, dmaResult);
         return dmaResult;

      case MEMORY:
         // we missed in all caches
         stats.pc_l1d.pc_ReadMisses.incr();
         stats.pc_l2d.pc_ReadMisses.incr();
         stats.pc_l3d.pc_ReadMisses.incr();
         // NB: fall-through to miss-handling code below
      }

      // we missed, so check remote caches for data
      dmaResult.remoteCommunicatedHappened = true;
      RemoteReadResponse rrr = performRemoteRead(access);

      MESIState newMesiState = MESIState.MESI_INVALID;
      if ( !rrr.providedData ) {
         // No Valid Read-Reply: Need to get this data from Memory
         memoryCyclesElapsed(MemorySystemConstants.MEMORY_ACCESS_LATENCY, dmaResult);

         /* NB: we always fetch from memory into Exclusive. */
         newMesiState = MESIState.MESI_EXCLUSIVE;

      } else if ( rrr.isShared ) {
         newMesiState = MESIState.MESI_SHARED;
         stats.pc_ReadRemoteHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.REMOTE_HIT_LATENCY, dmaResult);

      } else {
         // Valid Read-Reply From Modified/Exclusive
         newMesiState = MESIState.MESI_SHARED;
         stats.pc_ReadRemoteHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.REMOTE_HIT_LATENCY, dmaResult);
      }
      assert (newMesiState != MESIState.MESI_INVALID);

      line.changeStateTo(newMesiState);

      return dmaResult;

   } // end read()

   class RemoteReadResponse {
      boolean isShared     = false;
      boolean providedData = false;
   }

   private RemoteReadResponse performRemoteRead(final DataAccess access) {
      RemoteReadResponse rrr = new RemoteReadResponse();
      rrr.isShared = false;
      rrr.providedData = false;

      for ( Processor<Line> otherProc : allProcessors ) {
         if ( otherProc == this ) continue; // skip ourselves

         MemoryResponse<Line> resp = otherProc.L1cache.search(access.addr(),
                                                              params.remoteAccessesAffectLRU());
         Line otherLine = resp.lineHit;

         if ( resp.whereHit == CacheLevel.MEMORY ) continue;

         // found in a remote cache!

         switch ( otherLine.getState() ) {
         case MESI_EXCLUSIVE:
         case MESI_MODIFIED:
            otherLine.changeStateTo(MESIState.MESI_SHARED);
            rrr.providedData = true;
            return rrr;
         case MESI_SHARED:
            // everyone else will be in Shared state as well, so return now
            rrr.isShared = true;
            return rrr;
         case MESI_INVALID:
            // keep searching for other copies
            break;
         }

      } // done with other caches

      // this happens if everyone was MESI_INVALID
      return rrr;
   } // end readRemoteAction()

   /**
    * Make a write request.
    * 
    * @param access
    *           the memory access being made
    */
   public DataMemoryAccessResult write(final DataAccess access) {

      stats.pc_TotalDataWrites.incr();
      stats.pc_TotalMemoryAccesses.incr();

      DataMemoryAccessResult dmaResult = new DataMemoryAccessResult();

      if ( RadishSim.XASSERTS ) Verify.cacheExclusivity();

      MemoryResponse<Line> resp = L1cache.request(access.addr());
      Line myLine = resp.lineHit;

      if ( resp.whereHit != CacheLevel.MEMORY ) {

         switch ( resp.whereHit ) {
         case L1:
            stats.pc_l1d.pc_WriteHits.incr();
            memoryCyclesElapsed(MemorySystemConstants.L1_HIT_LATENCY, dmaResult);
            break;
         case L2:
            stats.pc_l1d.pc_WriteMisses.incr();
            stats.pc_l2d.pc_WriteHits.incr();
            memoryCyclesElapsed(MemorySystemConstants.L2_HIT_LATENCY, dmaResult);
            break;
         case L3:
            stats.pc_l1d.pc_WriteMisses.incr();
            stats.pc_l2d.pc_WriteMisses.incr();
            stats.pc_l3d.pc_WriteHits.incr();
            memoryCyclesElapsed(MemorySystemConstants.L3_HIT_LATENCY, dmaResult);
            break;
         case MEMORY:
            assert false;
         }

         // we either hit somewhere in our private cache(s) or we hit in the shared cache,
         // but we may not have write permissions

         switch ( myLine.getState() ) {
         case MESI_SHARED: // upgrade miss
            stats.pc_UpgradeMisses.incr();
            performRemoteWrite(access); // invalidate other copies
            memoryCyclesElapsed(MemorySystemConstants.REMOTE_HIT_LATENCY, dmaResult);

            myLine.changeStateTo(MESIState.MESI_MODIFIED);
            dmaResult.remoteCommunicatedHappened = true;
            return dmaResult;
         case MESI_EXCLUSIVE: // write hit
         case MESI_MODIFIED:
            myLine.changeStateTo(MESIState.MESI_MODIFIED);
            dmaResult.remoteCommunicatedHappened = false;
            return dmaResult;
         case MESI_INVALID:
            assert false;
         }

      }

      // we didn't have the line at all - need to check for remote copies
      stats.pc_l1d.pc_WriteMisses.incr();
      stats.pc_l2d.pc_WriteMisses.incr();
      stats.pc_l3d.pc_WriteMisses.incr();

      boolean invalidationsHappened = performRemoteWrite(access);

      if ( !invalidationsHappened ) { // nobody else had a copy of this line
         memoryCyclesElapsed(MemorySystemConstants.MEMORY_ACCESS_LATENCY, dmaResult);
      } else {
         stats.pc_WriteRemoteHits.incr();
         memoryCyclesElapsed(MemorySystemConstants.REMOTE_HIT_LATENCY, dmaResult);
      }

      myLine.changeStateTo(MESIState.MESI_MODIFIED);
      dmaResult.remoteCommunicatedHappened = true;
      return dmaResult;
   } // end write()

   /**
    * Invalidate remote copies of this line
    * 
    * @return true if any invalidations were performed
    */
   private boolean performRemoteWrite(final DataAccess access) {

      boolean performedInvalidation = false;
      for ( Processor<Line> otherProc : allProcessors ) {
         if ( otherProc == this ) continue; // skip ourselves

         MemoryResponse<Line> resp = otherProc.L1cache.search(access.addr(),
                                                              params.remoteAccessesAffectLRU());
         Line otherLine = resp.lineHit;

         if ( resp.whereHit == CacheLevel.MEMORY ) continue;

         // found a remote copy of this line

         switch ( otherLine.getState() ) {
         case MESI_MODIFIED:
         case MESI_EXCLUSIVE:
         case MESI_SHARED:
            otherLine.invalidate();
            performedInvalidation = true;
            // have to keep searching to find all Shared copies
            break;
         case MESI_INVALID:
            // keep searching for more copies
            break;
         }
      } // done with other caches

      return performedInvalidation;
   } // end writeRemoteAction()

   /**
    * @param syncop
    *           whether this is a source or a sink in the HB graph
    * @param validSource
    *           whether there is a valid source for this sink op (e.g. the first lock acquire has no
    *           "source")
    * @param syncObject
    *           the object being synchronized on
    */
   public void syncOp(SynchronizationOperationType syncop, boolean validSource, long syncObject) {

      switch ( this.params.simulationMode() ) {
      case BASELINE:
         break;
      case FASTTRACK:
         stats.pc_ExecDrivenCycleCount.incr(100);
         break;
      case RADISH:
         stats.pc_ExecDrivenCycleCount.incr(200);
         break;
      default:
         assert false;
      }

      // flush pending async sw checks (across all async cores) at synchronization points
      {
         long timeWhenAllChecksDone = 0;
         for ( CheckerProcessor<Line> checker : checkerProcs ) {
            long done = checker.flushChecks(this.id);
            timeWhenAllChecksDone = Math.max(timeWhenAllChecksDone, done);
         }

         if ( timeWhenAllChecksDone > stats.pc_ExecDrivenCycleCount.get() ) {
            stats.pc_ExecDrivenCycleCount.set(timeWhenAllChecksDone);
         }
      }

      switch ( syncop ) {
      case SOURCE:
      // this is a HB source (e.g. lock release), so record our current VC so that the sink can
      // read it
      {
         SyncObject sobj = vcOfSyncObject.get(syncObject);
         if ( sobj == null ) {
            // NB: a shallow copy of currentVectorClock is ok, since Epochs are immutable
            sobj = new SyncObject();
            sobj.vc = currentVectorClock.clone();
            sobj.timestamp = stats.pc_ExecDrivenCycleCount.get();
            vcOfSyncObject.put(syncObject, sobj);

         } else {
            Epoch[] clocks = sobj.vc;
            assert clocks != null;
            assert clocks.length == currentVectorClock.length;
            for ( int i = 0; i < currentVectorClock.length; i++ ) {
               clocks[i] = Epoch.Compare.max(currentVectorClock[i], clocks[i]);
            }
         }
      }

         stats.pc_SyncSources.incr();
         break;

      case SINK: // lock acquire
         stats.pc_SyncTotalSinks.incr();

         // pull up the VC of the source and merge it with ours
         SyncObject sobj = vcOfSyncObject.get(syncObject);

         if ( sobj == null ) {

            // are we e.g. the first acquirer of a lock?
            if ( !validSource ) {
               stats.pc_SyncSourcelessSinks.incr();
            }

            Epoch[] clocks = new Epoch[this.allProcessors.length];
            Arrays.fill(clocks, Epoch.EMPTY_EPOCH);
            
            sobj = new SyncObject();
            sobj.vc = clocks;
            sobj.timestamp = stats.pc_ExecDrivenCycleCount.get();
            vcOfSyncObject.put(syncObject, sobj);

            stats.pc_SyncUnmatchedSinks.incr();
            return;
         }
         Epoch[] sourceVC = sobj.vc;

         // merging VCs is an element-wise max()
         assert sourceVC.length == currentVectorClock.length;
         for ( int i = 0; i < currentVectorClock.length; i++ ) {
            currentVectorClock[i] = Epoch.Compare.max(currentVectorClock[i], sourceVC[i]);
         }
         // NB: sinks needn't increment our local clock
         
         stats.pc_ExecDrivenCycleCount.set(Math.max(stats.pc_ExecDrivenCycleCount.get(), sobj.timestamp));
         
         return;
      }

      /*
       * NB: technically, allocation should be delayed until an access in the new epoch occurs, to
       * avoid allocating an id for an epoch that contains no accesses. But this is unlikely to ever
       * be the case, so we just automatically allocate a new epoch here to simplify things.
       * 
       * With clock GC, we don't actually allocate any id's until an access actually needs them, but
       * we do check here that we *can* allocate a new id without rollover.
       * 
       * With the type-safety opt, this auto-allocation is somewhat less justified, as we only use
       * the allocation if there is a *byte* access within the epoch. It's still probably okay to be
       * conservative in this way, however, as most epochs probably have at least one byte access.
       */
      if ( epochOffsetManager.canAllocate() ) {
         epochOffsetManager.incrementCurrent();
         currentVectorClock[id.get()] = localEpochBase.add(epochOffsetManager.current());
         return;
      }

      // local epoch counters are about to roll over

      stats.pc_EpochRollovers.incr();
      stats.pc_ExecDrivenCycleCount.incr(params.epochRolloverCost());

      final RadishLine.EpochOffsetVisitor verifyEpochOffsets = new RadishLine.EpochOffsetVisitor() {
         @Override
         public void visit(final MemoryAccessType _, final MetadataByteAddress __,
                           final ByteMetadata md) {
            EpochOffset eo = md.epochOffset;
            if ( params.useClockGC() ) {
               if ( !params.useTypeSafetyOpt() || md.byteSized ) {
                  assert eo.untouched() || epochOffsetManager.isAllocated(eo);
               }
            } else {
               assert EpochOffset.Compare.lte(eo, epochOffsetManager.MAX_BASEOFFSET_VALUE());
            }
         }
      };

      HierarchicalCache.LineVisitor<Line> resetPrivateCacheEpochOffsets = new HierarchicalCache.LineVisitor<Line>() {
         @Override
         public void visit(Line line) {
            if ( !line.isMetadata() ) return;
            MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
            // only our metadata should be in our private caches
            assert mla.owner().equals(/* this Processor's */id);
            if ( RadishSim.XASSERTS ) line.visitLocalEpochOffsets(verifyEpochOffsets);
            evictMetadataToMemory(line, false, null);
            line.resetLocalEpochOffsets();
         }
      };

      HierarchicalCache.LineVisitor<Line> resetSharedCacheEpochOffsets = new HierarchicalCache.LineVisitor<Line>() {
         @Override
         public void visit(Line line) {
            if ( !line.isMetadata() ) return;
            MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
            // only flush *our* metadata lines
            if ( !mla.owner().equals(/* this Processor's */id) ) return;
            if ( RadishSim.XASSERTS ) line.visitLocalEpochOffsets(verifyEpochOffsets);
            evictMetadataToMemory(line, true, null);
            line.resetLocalEpochOffsets();
         }
      };

      // walk our private caches and reset the local epoch offsets
      L1cache.visitAllLines(resetPrivateCacheEpochOffsets);
      if ( L2cache != null ) {
         L2cache.visitAllLines(resetPrivateCacheEpochOffsets);
      }
      if ( L3cache != null ) {
         // flush our part of the L3 as well
         L3cache.visitAllLines(resetSharedCacheEpochOffsets);
      }

      assert epochOffsetManager.getRefCounts().isEmpty();

      // rebase
      EpochOffset beforeReset = epochOffsetManager.reset();
      Epoch newBase = localEpochBase.add(beforeReset);
      localEpochBase = new EpochBase(newBase.get());
      currentVectorClock[id.get()] = localEpochBase.add(epochOffsetManager.current());

   } // end syncOp()

   class CacheUsageSnapshot {
      long   timestamp             = 0;
      double invalidLines          = 0.0;
      double dataLines             = 0.0;
      double expandedMetadataLines = 0.0;
      double compactMetadataLines  = 0.0;
   }

   private void takeCacheUsageSnapshot() {
      // only p0 takes snapshots
      if ( id.get() != 0 ) return;

      lastCacheUsageSnapshot = stats.pc_ExecDrivenCycleCount.get();

      final CacheUsageSnapshot snapshot = new CacheUsageSnapshot();
      snapshot.timestamp = lastCacheUsageSnapshot;

      HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
         @Override
         public void visit(Line line) {
            if ( line.invalid() ) {
               snapshot.invalidLines++;

            } else if ( !line.isMetadata() ) {
               snapshot.dataLines++;

            } else if ( params.compactMetadata() && line.isCompact() ) {
               snapshot.compactMetadataLines++;

            } else {
               snapshot.expandedMetadataLines++;
            }
         }
      };

      for ( Processor<Line> p : allProcessors ) {
         p.L1cache.visitAllLines(lv);
         if ( p.L2cache != null ) {
            p.L2cache.visitAllLines(lv);
         }
      }
      if ( L3cache != null ) {
         L3cache.visitAllLines(lv);
      }

      // scale as proportion of total cache capacity
      double totalLines = snapshot.invalidLines + snapshot.dataLines
                          + snapshot.compactMetadataLines + snapshot.expandedMetadataLines;
      snapshot.invalidLines /= totalLines;
      snapshot.dataLines /= totalLines;
      snapshot.compactMetadataLines /= totalLines;
      snapshot.expandedMetadataLines /= totalLines;

      // store snapshot
      cacheUsageSnapshots.add(snapshot);
   }

   public void dumpCacheUsageSnapshots(Writer wr, String prefix, String suffix) throws IOException {
      for ( CacheUsageSnapshot s : cacheUsageSnapshots ) {
         StringWriter stat = new StringWriter();
         stat.write("'CacheUsageSnapshotTimestamp': " + s.timestamp + ",");
         stat.write("'InvalidLines': " + s.invalidLines + ",");
         stat.write("'DataLines': " + s.dataLines + ",");
         stat.write("'CompactMetadataLines': " + s.compactMetadataLines + ",");
         stat.write("'ExpandedMetadataLines': " + s.expandedMetadataLines);
         wr.write(prefix + stat.toString() + suffix);
      }
   }

   private class CompactMetadataAddresses {
      MetadataByteAddress bottom;
      MetadataByteAddress top;
      boolean             inTop;
   }

   /**
    * Calculate the bottom and top mba's corresponding to the given mba. Also returns whether the
    * given mba was in the top or the bottom part.
    */
   private CompactMetadataAddresses getCompactAddresses(MetadataByteAddress mba) {
      CompactMetadataAddresses cma = new CompactMetadataAddresses();

      final long bottomLineAddr = mba.get() & COMPACT_MD_ALIGNMENT_MASK();
      cma.bottom = MetadataByteAddress.Unscaled(bottomLineAddr, mba.owner());

      final long topLineAddr = bottomLineAddr
                               + (DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.EXPANDED) * EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      cma.top = MetadataByteAddress.Unscaled(topLineAddr, mba.owner());

      final long lineMba = mba.lineAddress().get();
      cma.inTop = (lineMba == topLineAddr);

      return cma;
   }

   /**
    * Find all lines that contain the address mba.
    * 
    * @return a map from Processor to that processor's Line containing metadata for mba. NB: the
    *         returned map may contain an entry for this (the calling Processor).
    */
   Map<Processor<Line>, Line> getMatchingMdLines(final MetadataByteAddress mba,
                                                 final boolean reorderSets) {
      Map<Processor<Line>, Line> res = new HashMap<Processor<Line>, Line>();

      for ( Processor<Line> proc : allProcessors ) {
         MetadataByteAddress otherProc_mba = new MetadataByteAddress(mba, proc.id);

         // this search covers if md is 1) expanded or 2) in the bottom half of a compact line
         MemoryResponse<Line> resp = proc.L1cache.ownedMdSearch(otherProc_mba, reorderSets);

         if ( params.compactMetadata() && resp.whereHit == CacheLevel.MEMORY ) {
            CompactMetadataAddresses cma = getCompactAddresses(otherProc_mba);
            // still need to search for md that's 3) in the top half of a compact line
            if ( cma.inTop ) {
               resp = proc.L1cache.ownedMdSearch(cma.bottom, reorderSets);
               if ( resp.whereHit == CacheLevel.MEMORY || resp.lineHit.isExpanded() ) {
                  continue;
               }
            }
         }
         if ( resp.whereHit == CacheLevel.MEMORY ) continue;

         Line line = resp.lineHit;
         // md should always be in Modified state
         assert line.isMetadata() && line.getState() == MESIState.MESI_MODIFIED;

         res.put(proc, line);
      }

      return res;
   }

   /**
    * Find all lines that contain metadata for any location within md. If md is a compact line,
    * multiple expanded lines for a given processor may be returned (hence the need to return a
    * multimap).
    * 
    * @return a multimap from a Processor to that processor's Line containing metadata for mba. NB:
    *         the returned map may contain an entry for this (the calling Processor).
    */
   Multimap<Processor<Line>, Line> getMatchingMdLines(final Line needle, final boolean reorderSets) {
      Multimap<Processor<Line>, Line> res = HashMultimap.create();
      MetadataByteAddress lineMba = new MetadataByteAddress(
                                                            (MetadataLineAddress) needle.lineAddress());

      if ( needle.isExpanded() ) {
         Map<Processor<Line>, Line> lines = getMatchingMdLines(lineMba, reorderSets);
         for ( Map.Entry<Processor<Line>, Line> me : lines.entrySet() ) {
            res.put(me.getKey(), me.getValue());
         }
         return res;
      }

      // md is compact; need to search for both top and bottom parts
      for ( Processor<Line> proc : allProcessors ) {
         MetadataByteAddress otherProc_mba = new MetadataByteAddress(lineMba, proc.id);
         CompactMetadataAddresses cma = getCompactAddresses(otherProc_mba);
         MetadataByteAddress mbas[] = { cma.bottom, cma.top };

         for ( int i = 0; i < mbas.length; i++ ) {
            MemoryResponse<Line> resp = proc.L1cache.ownedMdSearch(mbas[i], reorderSets);
            if ( resp.whereHit != CacheLevel.MEMORY ) {
               Line hit = resp.lineHit;
               // md should always be in Modified state
               assert hit.isMetadata() && hit.getState() == MESIState.MESI_MODIFIED;
               res.put(proc, hit);
            }
         }
      }

      return res;
   }

   /** compute the (additional) latency of this md access */
   private void computeMdAccessLatency(MemoryResponse<Line> resp, MetadataAccess access,
                                       DataMemoryAccessResult dataAccessResult) {
      int mdAccessLatency = 0;
      switch ( resp.whereHit ) {
      case L1:
         if ( access.read() ) {
            stats.pc_l1md.pc_ReadHits.incr();
         } else {
            stats.pc_l1md.pc_WriteHits.incr();
         }
         mdAccessLatency = MemorySystemConstants.L1_HIT_LATENCY;
         break;
      case L2:
         if ( access.read() ) {
            stats.pc_l1md.pc_ReadMisses.incr();
            stats.pc_l2md.pc_ReadHits.incr();
         } else {
            stats.pc_l1md.pc_WriteMisses.incr();
            stats.pc_l2md.pc_WriteHits.incr();
         }
         mdAccessLatency = MemorySystemConstants.L2_HIT_LATENCY;
         break;
      case L3:
         if ( access.read() ) {
            stats.pc_l1md.pc_ReadMisses.incr();
            stats.pc_l2md.pc_ReadMisses.incr();
            stats.pc_l3md.pc_ReadHits.incr();
         } else {
            stats.pc_l1md.pc_WriteMisses.incr();
            stats.pc_l2md.pc_WriteMisses.incr();
            stats.pc_l3md.pc_WriteHits.incr();
         }
         mdAccessLatency = MemorySystemConstants.L3_HIT_LATENCY;
         break;
      case MEMORY:
         if ( access.read() ) {
            stats.pc_l1md.pc_ReadMisses.incr();
            stats.pc_l2md.pc_ReadMisses.incr();
            stats.pc_l3md.pc_ReadMisses.incr();
         } else {
            stats.pc_l1md.pc_WriteMisses.incr();
            stats.pc_l2md.pc_WriteMisses.incr();
            stats.pc_l3md.pc_WriteMisses.incr();
         }
         mdAccessLatency = MemorySystemConstants.MEMORY_ACCESS_LATENCY;
      }
      // overlap as much latency as we can with the data access (which was already billed)
      int extraMdLatency = Math.max(mdAccessLatency, dataAccessResult.latency)
                           - dataAccessResult.latency;
      memoryCyclesElapsed(extraMdLatency, null);
   }

   private MemoryResponse<Line> fillExpandedLine(final MetadataByteAddress mba) {
      MemoryResponse<Line> resp = L1cache.request(mba, RadishLine.Compacted.EXPANDED);
      return resp;
   }

   /** Grow a compact line (already in-cache) into its companion line (not yet cached). */
   private MemoryResponse<Line> growLine(CompactMetadataAddresses cma) {

      // fetch the compact line
      MemoryResponse<Line> compactResp = L1cache.request(cma.bottom, RadishLine.Compacted.COMPACT);
      final Line compact = compactResp.lineHit;
      /*
       * compact line will hit if we're called from accessMetadata(), or may miss if we're called
       * from updateHwclocks().
       */

      // initialize compact line if necessary
      if ( compactResp.whereHit == CacheLevel.MEMORY ) {
         compact.changeStateTo(MESIState.MESI_MODIFIED);
         compact.setMetadata();
      }

      // fetch the companion line
      MemoryResponse<Line> companionResp = L1cache.request(cma.top, RadishLine.Compacted.EXPANDED);
      // companion shouldn't be cached, since we were in compact mode before
      assert companionResp.whereHit == CacheLevel.MEMORY;
      final Line companion = companionResp.lineHit;

      // initialize companion line
      companion.changeStateTo(MESIState.MESI_MODIFIED);
      companion.setMetadata();

      compact.expand(companion);

      // does access fall in bottom (former compact) or top (companion)?
      if ( cma.inTop ) {
         return companionResp;
      } else {
         return compactResp;
      }

   }

   /**
    * Update metadata for a read or write for the given `access' range.
    * 
    * @param access
    *           the bytes of metadata to access
    * @param dataAccessResult
    *           information about the corresponding data access
    * @return true if this access participates in a data race, false otherwise
    */
   public boolean accessMetadata(final MetadataAccess access,
                                 final DataMemoryAccessResult dataAccessResult) {
      if ( RadishSim.XASSERTS ) {
         Verify.verifyCacheInvariants();
      }

      final int snapshotFreq = params.cacheUsageSnapshotFrequencyInCycles();
      if ( snapshotFreq > 0
           && (stats.pc_ExecDrivenCycleCount.get() - lastCacheUsageSnapshot >= snapshotFreq) ) {
         takeCacheUsageSnapshot();
      }

      switch ( access.type() ) {
      case READ:
         stats.pc_TotalMetadataReads.incr();
         break;
      case WRITE:
         stats.pc_TotalMetadataWrites.incr();
         break;
      }

      boolean hit = false;
      MemoryResponse<Line> resp = null;
      final MetadataByteAddress expandedMba = access.addr();

      if ( !params.compactMetadata() ) {
         // compact md is disabled, so always request expanded form
         resp = L1cache.request(expandedMba, RadishLine.Compacted.EXPANDED);

      } else {
         // is our md present in compact form?
         final CompactMetadataAddresses cma = getCompactAddresses(expandedMba);

         final MetadataByteAddress compactMba = cma.bottom;
         final MetadataByteAddress companionMba = cma.top;
         final boolean accessInCompact = !cma.inTop;

         resp = L1cache.ownedMdSearch(compactMba, false);

         if ( resp.whereHit != CacheLevel.MEMORY ) {
            // hit with compact
            if ( resp.lineHit.isCompact() ) {
               if ( access.size() == 1 ) {
                  stats.pc_CompactMdExpansions.incr();
                  resp = growLine(cma);
               } else {
                  resp = L1cache.request(compactMba);
               }

            } else { // hit an expanded line, is it us or our companion?
               if ( accessInCompact ) { // hit our own line
                  resp = L1cache.request(compactMba);
               } else { // hit (expanded) companion, so our line needs to be expanded
                  // NB: our line may already be cached
                  resp = fillExpandedLine(expandedMba);
               }
            }

         } else { // missed with compact
            stats.pc_CompactMdCompanionChecks.incr();
            MemoryResponse<Line> companionResp = L1cache.ownedMdSearch(companionMba, false);

            if ( companionResp.whereHit != CacheLevel.MEMORY ) {
               // companion is expanded, so our line needs to be expanded too
               assert companionResp.lineHit.isExpanded();

               if ( !accessInCompact ) { // companion is our line
                  resp = L1cache.request(companionMba);
               } else { // not the line we're looking for
                  resp = fillExpandedLine(expandedMba);
                  assert resp.whereHit == CacheLevel.MEMORY;
               }

            } else { // didn't find the companion line
               if ( access.size() == 1 ) {
                  resp = fillExpandedLine(expandedMba);
                  assert resp.whereHit == CacheLevel.MEMORY;
               } else {
                  resp = L1cache.request(compactMba, RadishLine.Compacted.COMPACT);
                  assert resp.whereHit == CacheLevel.MEMORY;
               }
            }

         }
      }

      if ( RadishSim.XASSERTS ) {
         // verify evictions didn't mess anything up
         Verify.verifyCacheInvariants();
      }

      computeMdAccessLatency(resp, access, dataAccessResult);

      final Map<Processor<Line>, Line> lines;

      if ( resp.whereHit == CacheLevel.MEMORY ) {
         // missed in all caches, so we need to initialize line
         Line filled = resp.lineHit;
         filled.changeStateTo(MESIState.MESI_MODIFIED);
         filled.setMetadata();

         // NB: have to wait for filled line to be set valid
         lines = getMatchingMdLines(access.addr(), params.remoteAccessesAffectLRU());
         assert lines.containsValue(filled);

         MemoryAccessPermission[] perms;
         perms = Permissions.usingHardwareState(this, this.currentVectorClock,
                                                access.lineAddress(), lines);
         filled.assignLocalPerms(perms, access.lineAddress());
         if ( params.leafReads() == LeafReadConfiguration.LAZY_LEAF_READS && access.read() ) {
            clearLeafReads(access, lines);
         }

      } else { // cache hit
         hit = true;
         Line hitLine = resp.lineHit;

         lines = getMatchingMdLines(access.addr(), params.remoteAccessesAffectLRU());
         assert lines.containsValue(hitLine);

         assert hitLine.isMetadata() && hitLine.getState() == MESIState.MESI_MODIFIED;
         MetadataLineAddress mla = (MetadataLineAddress) hitLine.lineAddress();
         assert mla.owner() == this.id;

         // bring metadata back in sync, only when there was coherence traffic on the data line
         if ( dataAccessResult.remoteCommunicatedHappened ) {
            hitLine.assignLocalPerms(Permissions.usingHardwareState(this, this.currentVectorClock,
                                                                    access.lineAddress(), lines),
                                     access.lineAddress());
            if ( params.leafReads() == LeafReadConfiguration.LAZY_LEAF_READS && access.read() ) {
               clearLeafReads(access, lines);
            }
         }

      }

      // keep leaf reads fully precise "for free"
      if ( params.leafReads() == LeafReadConfiguration.PRECISE_LEAF_READS && access.read() ) {
         clearLeafReads(access, lines);
      }

      /*
       * Check if our read is totally-ordered with other reads (we can assume it's ordered with the
       * last write, otherwise there will be a race). This is just for instrumentation purposes.
       */
      if ( access.read() ) {
         boolean totallyOrderedRead = true;

         for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {

            // check hardware md
            for ( Map.Entry<Processor<Line>, Line> mentry : lines.entrySet() ) {
               Line l = mentry.getValue();
               Processor<Line> p = mentry.getKey();

               ByteMetadata md = l.getLocalEpochOffset(MemoryAccessType.READ, a.addr());
               Epoch readTime = p.localEpochBase.add(md.epochOffset);
               if ( md.epochOffset.touched()
                    && !Epoch.ha(this.currentVectorClock[p.id.get()], readTime) ) {
                  totallyOrderedRead = false;
               }
            }

            // check software md
            SoftwareVCPair vcp = this.globalMetaDataTable.get(a.addr().getUnowned());
            if ( vcp != null ) {
               Epoch[] reads = vcp.read;
               for ( int i = 0; i < reads.length; i++ ) {
                  if ( reads[i].touched() && !Epoch.ha(this.currentVectorClock[i], reads[i]) ) {
                     totallyOrderedRead = false;
                  }
               }
            }

         }

         if ( totallyOrderedRead ) stats.pc_TotallyOrderedMdReads.incr();
      }

      // now check local perms to see if we need to do any further checks
      boolean raceOccurred = false;
      Line l1line = resp.lineHit;
      if ( !l1line.accessWouldBeRaceFree(access) ) {
         // we have a local permissions violation, which means there might be a race
         if ( access.write() ) {
            stats.pc_LocalWriteViolatesLocalPerms.incr();
         } else {
            stats.pc_LocalReadViolatesLocalPerms.incr();
         }

         // update our local perms by doing a check with other caches
         if ( hit ) { // misses can piggyback these checks with the fill
            stats.pc_HwCacheChecks.incr();
            stats.pc_ExecDrivenCycleCount.incr(params.hwCheckCost());
         }

         raceOccurred = handleLocalPermissionsViolation(access, lines);
         // NB: if a sw check causes a compact line to grow, l1line is stale!
         l1line = lines.get(this);

         // piggy-back on hw race check
         if ( params.leafReads() == LeafReadConfiguration.LAZY_LEAF_READS && access.read() ) {
            clearLeafReads(access, lines);
         }

         // we always get read perms; on a write we get write perms, too
         l1line.setLocalPerms(access.write() ? MemoryAccessPermission.WRITE_OK
                                            : MemoryAccessPermission.READ_OK, access);
      }

      Verify.singleWriterInvariant(access, lines);

      /*
       * NB: In an actual implementation, we would only reset metadata on coherence activity. To
       * simplify the simulator design, we reset metadata on every (even purely local) write. This
       * simplification allows us to enforce the invariant that the metadata for only a single
       * writer is valid at any one time.
       * 
       * In reality, 2 processors can have metadata for a given byte at one time, though only one of
       * those processor's data will be valid. Consider: p0.write(1), p1.write(0) to get the line in
       * Modified state, then p1.write(1). There's no coherence traffic on p1's second write, so p0
       * still has md for byte 1, but p0's md is stale. On the next coherence event, p0's md must
       * first be invalidated before other vc calculations take place.
       */

      if ( /* remoteCommunicationHappened && */access.write() ) {
         resetMetadata(access, false, lines);
      }

      final boolean clobberReads = true; // writes clobber reads
      List<EpochOffset> freed = l1line.recordLocalAccess(access, epochOffsetManager.current(),
                                                         params.useTypeSafetyOpt(), clobberReads);
      if ( params.useClockGC() ) {
         for ( EpochOffset eo : freed ) {
            assert eo.touched();
            epochOffsetManager.decrRefCount(eo);
         }
         // update ref counts for all bytes in this access
         for ( int i = 0; i < access.size(); i++ ) {
            if ( !params.useTypeSafetyOpt() || access.size() == 1 ) {
               epochOffsetManager.incrRefCount(epochOffsetManager.current());
            }
         }
      }

      Verify.singleWriterInvariant(access, lines);

      // NB: hw/sw race checks may have already set in-hw status, but it's safe to set it again

      // writes set in-hw status
      if ( access.write() ) {
         setEverythingInHardwareNow(access, false);
      }

      // System.out.println( L1cache.toString() );
      if ( RadishSim.XASSERTS ) {
         Verify.verifyCacheInvariants();
      }

      return raceOccurred;

   } // end accessMetadata()

   /**
    * Do the necessary bookkeeping for evicted, which contains metadata and is about to be evicted
    * from the cache hierarchy.
    * 
    * @param evicted
    *           the line being evicted to memory
    * @param lineInLlc
    *           whether evicted is currently resident in the shared LLC
    * @param incoming
    *           The line entering the set from which evicted is being ejected. Null if we are called
    *           during a cache flush.
    */
   private void evictMetadataToMemory(Line evicted, boolean lineInLlc, Line incoming) {
      assert evicted.isMetadata() && evicted.getState() == MESIState.MESI_MODIFIED;

      // line should be in LLC, and not anywhere else
      if ( RadishSim.XASSERTS && lineInLlc ) Verify.lineIsOnlyInLLC(evicted);

      // count the program variables that were in this line
      // variablesInLine.put(evicted.lineAddress(), evicted.numLocalAccesses());

      // find this line's owner
      MetadataLineAddress mla = (MetadataLineAddress) evicted.lineAddress();
      final CpuId owner = mla.owner();
      final Processor<Line> evictionOwnerProc = allProcessors[owner.get()];

      /*
       * Are there other processors with md for locations contained within evicted? If not, we are
       * evicting the last version of this md and we need to downgrade in-hw status accordingly.
       * 
       * NB: this is an artifact of our centralized in-hw status bookkeeping. With distributed
       * bookkeeping, the status would naturally degrade when the last version was evicted.
       */

      final Multimap<Processor<Line>, Line> lines = getMatchingMdLines(evicted, false);

      // exclude evicted from ihs considerations
      boolean found = lines.remove(evictionOwnerProc, evicted);
      assert found;
      assert !lines.containsKey(evictionOwnerProc);

      // may need to take incoming into account for ihs calculations
      boolean foundOtherProcWithMetadata = false;
      if ( incoming != null && incoming.isMetadata() ) {

         if ( evicted.containsUnowned(new MetadataByteAddress(
                                                              (MetadataLineAddress) incoming.lineAddress()))
              || incoming.containsUnowned(new MetadataByteAddress(
                                                                  (MetadataLineAddress) evicted.lineAddress())) ) {
            foundOtherProcWithMetadata = true;

            CpuId incomingOwner = ((MetadataLineAddress) incoming.lineAddress()).owner();
            Processor<Line> p = this.allProcessors[incomingOwner.get()];
            lines.put(p, incoming);
         }
      }

      if ( !foundOtherProcWithMetadata ) foundOtherProcWithMetadata = !lines.isEmpty();

      final boolean evictingLastMetadata = !foundOtherProcWithMetadata;
      final boolean[] evictedValidMd = { false };
      final boolean[] evictedFalseLeafRead = { false };

      RadishLine.EpochOffsetVisitor evictValidMetadata = new RadishLine.EpochOffsetVisitor() {
         @Override
         public void visit(final MemoryAccessType mtype, final MetadataByteAddress mba,
                           final ByteMetadata evictedByteMd) {

            /*
             * Clear in-hw status if we're evicting the last metadata.
             * 
             * NB: we have to clear ihs even if we don't have valid md for this byte, since on a
             * previous sw check we may have set ihs for the whole line. This is an artifact of our
             * centralized ihs management.
             */
            UnownedMetadataByteAddress umba = mba.getUnowned();
            if ( evictingLastMetadata ) inHardwareStatus.remove(umba);

            // no valid metadata for this byte
            EpochOffset eo = evictedByteMd.epochOffset;
            if ( eo.untouched() ) return;

            if ( evictionOwnerProc.params.useClockGC() ) {
               if ( !params.useTypeSafetyOpt() || evictedByteMd.byteSized ) {
                  // decrement ref counts for outgoing epoch offsets
                  evictionOwnerProc.epochOffsetManager.decrRefCount(eo);
               }
            }

            if ( mtype == MemoryAccessType.READ && !evictedByteMd.leafRead ) {
               // the eviction of non-leaf reads doesn't degrade ihs or affect sw md
               return;
            }

            // at this point, we are evicting valid metadata
            evictedValidMd[0] = true;

            // record evicted metadata in the MemoryTable
            if ( mtype == MemoryAccessType.WRITE
                 || (mtype == MemoryAccessType.READ && evictedByteMd.leafRead) ) {
               Epoch e = evictionOwnerProc.localEpochBase.add(eo);
               assert e.touched();
               globalMetaDataTable.update(owner, mba.getUnowned(), mtype, e,
                                          evictedByteMd.byteSized);
            }

            // downgrade in-hw status
            InHardwareStatus ihs = inHardwareStatus.get(umba);

            // ihs already maximally downgraded
            if ( ihs == null || evictingLastMetadata ) return;

            switch ( mtype ) {
            case READ:
               // only the eviction of leaf reads causes ihs to degrade
               switch ( ihs ) {
               case EVERYTHING:
                  Epoch evictedRead = evictionOwnerProc.localEpochBase.add(eo);
                  evictedFalseLeafRead[0] |= isNonLeafRead(evictionOwnerProc, evictedRead, mba,
                                                           lines);

                  // downgrade to LAST_WRITE iff there is a last write in hardware
                  boolean lastWriteInHw = false;
                  for ( Map.Entry<Processor<Line>, Line> me : lines.entries() ) {
                     Line l = me.getValue();
                     if ( l.containsUnowned(mba) ) {
                        ByteMetadata b = l.getLocalEpochOffset(MemoryAccessType.WRITE, mba);
                        if ( b.epochOffset.touched() ) {
                           lastWriteInHw = true;
                           break;
                        }
                     }
                  }

                  if ( lastWriteInHw ) {
                     inHardwareStatus.put(umba, InHardwareStatus.LAST_WRITE);
                  } else {
                     inHardwareStatus.remove(umba);
                  }

                  break;
               case ALL_LEAF_READS:
                  inHardwareStatus.remove(umba);
                  break;
               case LAST_WRITE: // nop
                  break;
               }
               break;
            case WRITE:
               switch ( ihs ) {
               case EVERYTHING:

                  // downgrade to ALL_LEAF_READS iff there is a leaf read in hw
                  boolean leafReadInHw = false;
                  for ( Map.Entry<Processor<Line>, Line> me : lines.entries() ) {
                     Line l = me.getValue();
                     if ( l.containsUnowned(mba) ) {
                        ByteMetadata b = l.getLocalEpochOffset(MemoryAccessType.READ, mba);
                        if ( b.epochOffset.touched() && b.leafRead ) {
                           leafReadInHw = true;
                           break;
                        }
                     }
                  }

                  if ( leafReadInHw ) {
                     inHardwareStatus.put(umba, InHardwareStatus.ALL_LEAF_READS);
                  } else {
                     inHardwareStatus.remove(umba);
                  }
                  break;
               case LAST_WRITE:
                  inHardwareStatus.remove(umba);
                  break;
               case ALL_LEAF_READS:
                  assert false : "the valid write was already evicted";
               }
               break;
            }
         }
      };
      evicted.visitLocalEpochOffsets(evictValidMetadata);

      if ( evictedValidMd[0] ) evictionOwnerProc.stats.pc_MetadataEvictions.incr();
      if ( evictedFalseLeafRead[0] ) evictionOwnerProc.stats.pc_EvictedFalseLeafReads.incr();

      evicted.clearMetadata();
      evicted.invalidate();

   } // end evictMetadataToMemory()

   private boolean isNonLeafRead(Processor<Line> targetProc, Epoch targetRead,
                                 MetadataByteAddress mba, Multimap<Processor<Line>, Line> lines) {
      for ( Map.Entry<Processor<Line>, Line> other : lines.entries() ) {
         Processor<Line> otherProc = other.getKey();
         Line otherLine = other.getValue();

         if ( otherProc == targetProc ) continue;
         if ( !otherLine.containsUnowned(mba) ) continue;

         /*
          * In general, we want to know if an evicted read er hb some in-hw read rh by processor
          * prh. It is, however, impossible to know what happened-before rh if rh is in a previous
          * epoch, so we can only approximate this information by checking if rh is in prh's current
          * epoch. We can identify some falsely-identified leaf reads in this manner.
          */
         ByteMetadata otherByteMd = otherLine.getLocalEpochOffset(MemoryAccessType.READ, mba);
         EpochOffset otherOffset = otherByteMd.epochOffset;
         if ( otherOffset.touched() ) {
            if ( otherOffset.equals(otherProc.epochOffsetManager.current())
                 && Epoch.ha(otherProc.currentVectorClock[targetProc.id.get()], targetRead) ) {
               // is definitely a non-leaf read
               return true;
            }

         }
      }
      // could be a valid leaf read
      return false;
   }

   /**
    * Handle the local perms violation by 1) trying to resolve the violation in hardware if possible
    * and 2) failing over to software if not.
    * 
    * @return whether this access participates in a data race
    */
   private boolean handleLocalPermissionsViolation(final MetadataAccess access,
                                                   Map<Processor<Line>, Line> lines) {
      boolean wasDataRace = false;

      // check whether we have sufficient md in hardware to resolve this conflict
      stats.pc_HwSufficiencyChecks.incr();
      switch ( hardwareRaceCheck(access, lines) ) {

      case HW_CHECK_RACE_FREE:
         switch ( access.type() ) {
         case WRITE: // writes blow away everything
            resetMetadata(access, false, lines);
            break;
         case READ:
            // Reads don't invalidate any epochs, or change in-hw status.

            // Set local perms for everyone. NB: this is a bit redundant with
            // accessMetadata(), which sets our perms for the access bytes.
            recalculateAllPerms(access, lines);
            break;
         }
         break;

      case HW_CHECK_FOUND_RACE:
         // races blow away all previous accesses, because it's too hard to recover consistency
         stats.pc_DataRaces.incr();
         wasDataRace = true;
         resetMetadata(access, true, lines);
         break;

      case HW_CHECK_UNSURE: // have to go to software
         stats.pc_SwVcChecksTotal.incr();

         MetadataAccess checkedAccess = access;
         MetadataAccess updatedAccess = access;
         if ( params.useWholeLineSoftwareChecks() ) {
            MetadataByteAddress lineStart = new MetadataByteAddress(access.lineAddress());
            // NB: update only the expanded access, to be conservative
            updatedAccess = new MetadataAccess(
                                               access.type(),
                                               lineStart,
                                               MemorySystemConstants.DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.EXPANDED));
            /*
             * Integer numVars = variablesInLine.get(access.lineAddress()); if ( numVars != null ) {
             * this.stats.pc_VarsCheckedInSoftwareChecks.incr(numVars);
             * this.stats.pc_SoftwareChecksWithVarInfo.incr(); }
             */
         }
         SwRaceCheckResult res = softwareRaceCheck(updatedAccess, checkedAccess, lines);
         stats.pc_ExecDrivenCycleCount.incr(this.params.swCheckCost()); // fixed cost

         // variable cost: sync/async processing of actual FastTrack accesses
         if ( params.simulateSoftwareChecks() ) { // NB: simulateSoftwareChecks() is only ever false in testing
            if ( params.asyncProcessingCores() == 0 ) {
               machine.enqueueAccesses(this.id, res.swCheckAccesses);
               stats.pc_ExecDrivenCycleCount.incr(MemorySystemConstants.FASTTRACK_CHECK_INSN_COST);

            } else { // we can potentially run this check async
               final long now = this.stats.pc_ExecDrivenCycleCount.get();

               List<CheckerProcessor<Line>> potentialCheckers = new LinkedList<CheckerProcessor<Line>>();
               for ( CheckerProcessor<Line> cp : this.checkerProcs ) {
                  // estimate if this checker core can get things done faster than we can
                  if ( cp.currentTime() <= (now + MemorySystemConstants.APPROX_SW_CHECK_COST) ) {
                     potentialCheckers.add(cp);
                  }
               }

               boolean ranAsync = false;
               switch ( potentialCheckers.size() ) {
               case 0:
                  // no available checker cores; perform check synchronously
                  machine.enqueueAccesses(this.id, res.swCheckAccesses);
                  stats.pc_ExecDrivenCycleCount.incr(MemorySystemConstants.FASTTRACK_CHECK_INSN_COST);
                  break;
               // 1+ potential checkers: check can run async
               case 1:
                  ranAsync = true;
                  potentialCheckers.get(0).scheduleCheck(this.id, now, res.swCheckAccesses);
                  break;
               default:
                  ranAsync = true;
                  // schedule check on a random available checker core               
                  int i = rng.nextInt(potentialCheckers.size());
                  potentialCheckers.get(i).scheduleCheck(this.id, now, res.swCheckAccesses);
               }

               if ( ranAsync ) { // check ran asynchronously
                  switch ( access.type() ) {
                  case READ:
                     stats.pc_SwVcChecksAsync_DueToReads.incr();
                     stats.pc_SwVcChecksTotal_DueToReads.incr();
                     break;
                  case WRITE:
                     stats.pc_SwVcChecksAsync_DueToWrites.incr();
                     stats.pc_SwVcChecksTotal_DueToWrites.incr();
                     break;
                  }

               } else { // check was synchronous
                  switch ( access.type() ) {
                  case READ:
                     stats.pc_SwVcChecksSync_DueToReads.incr();
                     stats.pc_SwVcChecksTotal_DueToReads.incr();
                     break;
                  case WRITE:
                     stats.pc_SwVcChecksSync_DueToWrites.incr();
                     stats.pc_SwVcChecksTotal_DueToWrites.incr();
                     break;
                  }
               }
            }
         }

         // set in-hw status
         int i = 0;
         for ( MetadataAccess a = new MetadataAccess(updatedAccess); a.keepGoing(); a.incr(), i++ ) {
            if ( res.newInHardwareStatus[i] == null ) {
               this.inHardwareStatus.remove(a.addr().getUnowned());
            } else {
               this.inHardwareStatus.put(a.addr().getUnowned(), res.newInHardwareStatus[i]);
               if ( res.newInHardwareStatus[i] == InHardwareStatus.EVERYTHING ) {
                  this.globalMetaDataTable.remove(a.addr().getUnowned());
               }
            }
         }

         // Set local perms for everyone. NB: this is a bit redundant with
         // accessMetadata(), which sets our perms for the access bytes.
         recalculateAllPerms(access, lines);

         if ( res.foundRace || access.write() ) {
            // races and writes blow away all previous accesses
            resetMetadata(access, true, lines);
         }
         if ( res.foundRace ) {
            wasDataRace |= res.foundRace;
            this.stats.pc_DataRaces.incr();
         }

         break;
      }

      return wasDataRace;
   }

   /**
    * Reset hardware state so that only access is recorded. The metadata for all other accesses is
    * cleared, for all remote processors. This is typically done as a result of a write or a race.
    * 
    * @param clearAllMd
    *           If true, clear md for all processors (including the calling processor), and software
    *           md as well. Otherwise, clear md for all remote processors.
    */
   private void resetMetadata(final MetadataAccess access, boolean clearAllMd,
                              Map<Processor<Line>, Line> lines) {
      for ( Map.Entry<Processor<Line>, Line> me : lines.entrySet() ) {
         Line line = me.getValue();
         Processor<Line> proc = me.getKey();

         if ( !clearAllMd && proc == this ) continue; // skip ourselves
         // NB: our epochs, our local perms are set by accessMetadata()

         // invalidate all remote epochs
         List<EpochOffset> freedEpochOffsets = line.invalidateLocalEpochs(access,
                                                                          params.useTypeSafetyOpt());
         if ( params.useClockGC() ) {
            for ( EpochOffset eo : freedEpochOffsets ) {
               assert eo.touched();
               proc.epochOffsetManager.decrRefCount(eo);
            }
         }

         // set local perms for otherProc
         MemoryAccessPermission otherPerm = access.write() ? MemoryAccessPermission.NOTHING_OK
                                                          : MemoryAccessPermission.READ_OK;
         line.setLocalPerms(otherPerm, access);
      }

      // we're the only access now; accessMetadata() will redundantly do this on writes
      setEverythingInHardwareNow(access, clearAllMd);
   }

   /** Downgrade all reads where the caller is totally-ordered with the reader. */
   private void clearLeafReads(final MetadataAccess access, Map<Processor<Line>, Line> lines) {
      assert params.leafReads() != LeafReadConfiguration.NO_LEAF_READS;

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         for ( Map.Entry<Processor<Line>, Line> mentry : lines.entrySet() ) {
            Line remoteLine = mentry.getValue();
            Processor<Line> remoteProc = mentry.getKey();

            // skip ourselves, since we are definitely a leaf read
            if ( remoteProc == this ) continue;

            ByteMetadata remoteMd = remoteLine.getLocalEpochOffset(MemoryAccessType.READ, a.addr());
            Epoch remoteRead = remoteProc.localEpochBase.add(remoteMd.epochOffset);
            if ( Epoch.ha(this.currentVectorClock[remoteProc.id.get()], remoteRead) ) {
               // we are well-ordered with this read, so defoliate it
               remoteLine.invalidateLeafReads(a.addr());
            }
         }
      }
   }

   /** Recalculate each processor's local perms, for the cache lines provided */
   private void recalculateAllPerms(MetadataAccess access, Map<Processor<Line>, Line> lines) {
      for ( Map.Entry<Processor<Line>, Line> me : lines.entrySet() ) {
         Line line = me.getValue();
         Processor<Line> proc = me.getKey();
         MemoryAccessPermission[] perms;
         perms = Permissions.usingHardwareState(proc, proc.currentVectorClock,
                                                access.lineAddress(), lines);
         line.assignLocalPerms(perms, access.lineAddress());
      }
   }

   enum HwRaceCheckResult {
      HW_CHECK_UNSURE, HW_CHECK_RACE_FREE, HW_CHECK_FOUND_RACE
   };

   /**
    * Check if there is sufficient state in hardware to precisely discharge the data-race-check on
    * this access.
    * 
    * @return true if hw state was sufficient, false if software state must be consulted
    */
   private HwRaceCheckResult hardwareRaceCheck(final MetadataAccess access,
                                               Map<Processor<Line>, Line> lines) {
      assert access.owner().equals(this.id);
      boolean accessIsRaceFree = true;

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {

         MemoryAccessPermission perm = Permissions.usingHardwareState(this,
                                                                      this.currentVectorClock,
                                                                      a.addr(), lines);
         switch ( access.type() ) {
         case READ:
            if ( MemoryAccessPermission.Compare.gte(perm, MemoryAccessPermission.READ_OK) ) {
               continue;
            } else accessIsRaceFree = false;
            break;
         case WRITE:
            if ( MemoryAccessPermission.Compare.gte(perm, MemoryAccessPermission.WRITE_OK) ) {
               continue;
            } else accessIsRaceFree = false;
            break;
         }
         // we have insufficient perms

         InHardwareStatus ihs = this.inHardwareStatus.get(a.addr().getUnowned());
         if ( ihs == null ) {
            // may exist some reads in hw, but (we are writing ^ we don't ha any reads)
            return HwRaceCheckResult.HW_CHECK_UNSURE;
         }

         switch ( ihs ) {
         case EVERYTHING:
            continue;
         case LAST_WRITE:
            if ( access.read() ) { // last write is sufficient for reads
               continue;
            } else { // access.write()
               // need to know about readers that are in sw
               return HwRaceCheckResult.HW_CHECK_UNSURE;
            }
         case ALL_LEAF_READS:
            if ( access.write() ) { // these reads ha the last write, so we can check a write
               continue;
            } else { // access.read()
               // we don't ha any reader (otherwise we'd pass the perms check above), so we have to
               // check against last writer, who is in sw
               return HwRaceCheckResult.HW_CHECK_UNSURE;
            }
         }

      } // end access iteration

      // if we make it here, we have precise information for all bytes in the access
      return accessIsRaceFree ? HwRaceCheckResult.HW_CHECK_RACE_FREE
                             : HwRaceCheckResult.HW_CHECK_FOUND_RACE;

   } // end existsSufficientMetadata()

   class SwRaceCheckResult {
      boolean            foundRace;
      InHardwareStatus[] newInHardwareStatus;
      List<DataAccess>   swCheckAccesses;
   }

   /** Perform a FastTrack race check for the specified access, updating software metadata as well. */
   public SwRaceCheckResult fastTrackRaceCheck(final MetadataAccess checkedAccess) {
      SwRaceCheckResult res = new SwRaceCheckResult();
      res.foundRace = false;
      res.newInHardwareStatus = null;
      res.swCheckAccesses = null;

      // NB: we only do a FT check for the first byte of the access; we assume all
      // accesses are the same.
      MetadataByteAddress mba = checkedAccess.addr();
      SoftwareVCPair swClocks = this.globalMetaDataTable.get(mba.getUnowned());
      res.swCheckAccesses = SoftwareMetadata.doFastTrackCheck(swClocks, mba.getUnowned(),
                                                              checkedAccess.type(), this.id,
                                                              this.currentVectorClock);

      // update software metadata
      this.globalMetaDataTable.update(this.id, mba.getUnowned(), checkedAccess.type(),
                                      this.currentVectorClock[this.id.get()],
                                      checkedAccess.size() == 1);

      return res;
   }

   /**
    * Perform a software race check.
    * 
    * @param updatedAccess
    *           the bytes whose md are updated
    * @param checkedAccess
    *           the bytes (contained in updatedAccess) that are actually checked for races
    * @return whether any bytes in checkedAccess are involved in a race
    */
   private SwRaceCheckResult softwareRaceCheck(final MetadataAccess updatedAccess,
                                               final MetadataAccess checkedAccess,
                                               Map<Processor<Line>, Line> lines) {

      SwRaceCheckResult res = new SwRaceCheckResult();
      res.foundRace = false;
      res.newInHardwareStatus = new InHardwareStatus[updatedAccess.size()];
      res.swCheckAccesses = null;

      boolean softwareHadNewerClock = false;
      boolean softwareClocksEmpty = true;

      int ihsIndex = 0;
      Line aLine = null;
      if (!lines.isEmpty()) aLine = lines.values().iterator().next(); 
      for ( MetadataAccess a = new MetadataAccess(updatedAccess); a.keepGoing(); a.incr(), ihsIndex++ ) {

         MetadataByteAddress mba = a.addr();
         HardwareVCPair hwClocks = Permissions.getHardwareReadWriteClocks(mba, lines,
                                                                          this.allProcessors.length);
         SoftwareVCPair swClocks = this.globalMetaDataTable.get(mba.getUnowned());
         if ( swClocks != null ) {
            softwareClocksEmpty = false;
            assert hwClocks.read.length == swClocks.read.length;
         }
         
         // do a FT check for the first byte of each program var
         if ( mba.get() == checkedAccess.addr().get() || aLine.startOfProgramVariable(mba) ) {
             List<DataAccess> da;
             da = SoftwareMetadata.doFastTrackCheck(swClocks, mba.getUnowned(),
                                                    a.type(), this.id,
                                                    this.currentVectorClock);
             if ( res.swCheckAccesses == null ) res.swCheckAccesses = da;
             else res.swCheckAccesses.addAll(da);
         }

         MergingResult mr = Permissions.mergeHwAndSwClocks(hwClocks, swClocks);
         HardwareVCPair freshClocks = mr.mergedClocks;
         softwareHadNewerClock |= mr.softwareHadNewerClock;

         // update hw clocks with values from sw
         if ( swClocks != null ) {
            // NB: updating write md will clobber read md, so perform write update first!
            boolean hwUpdated;
            Processor<Line> swProc = null;
            if ( swClocks.hasWrite() ) swProc = this.allProcessors[swClocks.getWriter().get()];
            hwUpdated = updateHwClocks(MemoryAccessType.WRITE, swProc, hwClocks.getWrite(),
                                       swClocks.getWrite(), mba, swClocks.byteSized, lines);
            if ( hwUpdated ) {
               hwClocks.setWrite(swClocks);
               swClocks.clearWrite();
            }

            for ( int i = 0; i < hwClocks.read.length; i++ ) {
               hwUpdated = updateHwClocks(MemoryAccessType.READ, this.allProcessors[i],
                                          hwClocks.read[i], swClocks.read[i], mba,
                                          swClocks.byteSized, lines);
               if ( hwUpdated ) {
                  hwClocks.read[i] = swClocks.read[i];
                  swClocks.read[i] = Epoch.EMPTY_EPOCH;
               }
            }
         }

         // update leafReads
         switch ( params.leafReads() ) {
         case PRECISE_LEAF_READS:
         case LAZY_LEAF_READS: {
            Epoch[] reads = freshClocks.read;
            for ( int i = 0; i < reads.length; i++ ) {
               if ( reads[i].untouched() || !freshClocks.leafRead[i] ) continue;

               // check if exists proc j that has a read ordered after i's supposedly-leaf read
               for ( int j = 0; j < reads.length; j++ ) {
                  if ( j == i ) continue;
                  Processor<Line> jProc = this.allProcessors[j];
                  Epoch jCurrentEpoch = jProc.localEpochBase.add(jProc.epochOffsetManager.current());
                  if ( reads[j].equals(jCurrentEpoch)
                       && Epoch.ha(jProc.currentVectorClock[i], reads[i]) ) {
                     // i's read is not a leaf read
                     freshClocks.leafRead[i] = false;
                     // update hw, too
                     if ( hwClocks.read[i].touched() ) {
                        Line l = lines.get(this.allProcessors[i]);
                        l.invalidateLeafReads(mba);
                     }
                  }

               } // end for j

            } // end for i
         }
            break;
         case NO_LEAF_READS:
            break;
         }

         // calculate strongest valid in-hw status
         boolean lastWriteInHw = hwClocks.hasWrite();

         boolean validLeafReadInSw = false;
         boolean leafReadInHw = false;
         for ( int i = 0; i < freshClocks.read.length; i++ ) {
            if ( freshClocks.read[i].touched() && freshClocks.leafRead[i]
                 && hwClocks.read[i].untouched() ) {
               validLeafReadInSw = true;
            }
            if ( hwClocks.read[i].touched() && hwClocks.leafRead[i] ) {
               leafReadInHw = true;
            }
         }
         // the currently executing read is a leaf read, but its md isn't set yet
         if ( checkedAccess.contains(mba) && a.type() == MemoryAccessType.READ ) {
            leafReadInHw = true;
         }

         if ( (lastWriteInHw || !freshClocks.hasWrite()) && !validLeafReadInSw ) {
            res.newInHardwareStatus[ihsIndex] = InHardwareStatus.EVERYTHING;
         } else if ( leafReadInHw && !lastWriteInHw && !validLeafReadInSw ) {
            res.newInHardwareStatus[ihsIndex] = InHardwareStatus.ALL_LEAF_READS;
         } else if ( lastWriteInHw && validLeafReadInSw ) {
            res.newInHardwareStatus[ihsIndex] = InHardwareStatus.LAST_WRITE;
         }

         // check whether a race occurred, only for the bytes of the actual access
         if ( res.foundRace ) continue; // skip redundant race checks
         if ( checkedAccess.contains(mba) ) {
            MemoryAccessPermission ourPerms;
            ourPerms = Permissions.calculate(freshClocks, this.currentVectorClock);
            switch ( a.type() ) {
            case READ:
               if ( MemoryAccessPermission.Compare.lt(ourPerms, MemoryAccessPermission.READ_OK) ) {
                  res.foundRace = true;
                  // don't return yet as we need to initialize in-hw status for the whole access
               }
               break;
            case WRITE:
               if ( MemoryAccessPermission.Compare.lt(ourPerms, MemoryAccessPermission.WRITE_OK) ) {
                  res.foundRace = true;
                  // don't return yet as we need to initialize in-hw status for the whole access
               }
               break;
            }
         }

      } // end access iteration

      if ( !softwareHadNewerClock ) {
         if ( softwareClocksEmpty ) {
            stats.pc_UselessSwChecks_SwEmpty_Total.incr();
            switch ( checkedAccess.type() ) {
            case READ:
               stats.pc_UselessSwChecks_SwEmpty_Read.incr();
               break;
            case WRITE:
               stats.pc_UselessSwChecks_SwEmpty_Write.incr();
               break;
            }

         } else {
            stats.pc_UselessSwChecks_SwStale_Total.incr();
            switch ( checkedAccess.type() ) {
            case READ:
               stats.pc_UselessSwChecks_SwStale_Read.incr();
               break;
            case WRITE:
               stats.pc_UselessSwChecks_SwStale_Write.incr();
               break;
            }
         }
      }

      return res;
   }

   /**
    * Update the entries in the hardware vc (hwc) where software has a more up-to-date value.
    * 
    * @return true if hardware should be updated with software's value
    */
   private boolean updateHwClocks(final MemoryAccessType mtype, final Processor<Line> swProc,
                                  final Epoch hwc, final Epoch swc, final MetadataByteAddress mba,
                                  final boolean byteSized, Map<Processor<Line>, Line> lines) {
      List<EpochOffset> toFree = new LinkedList<EpochOffset>();

      if ( hwc.touched() || swc.untouched() ) { return false; }
      // hwc.untouched() && swc.touched()

      Line line = lines.get(swProc);
      if ( line == null ) return false;

      EpochOffset newEpochOffset = swc.subtract(swProc.localEpochBase);
      boolean doUpdate = false;
      if ( params.useClockGC() ) {
         // NB: with clockGC, we can store negative offsets and therefore values <= epoch base

         if ( !params.useTypeSafetyOpt() || byteSized ) {
            // try to allocate and overwrite
            if ( swProc.epochOffsetManager.isAllocated(newEpochOffset)
                 || swProc.epochOffsetManager.canAllocate() ) {
               swProc.epochOffsetManager.incrRefCount(newEpochOffset);
               doUpdate = true;
            }

         } else { // params.useTypeSafetyOpt && !byteSized
            // we can always overwrite
            doUpdate = true;
         }

      } else { // base+offset encoding
         // NB: since 0 is an invalid epoch offset (it's the untouched epoch), we can't encode
         // anything <= proc's epoch base
         Epoch procMinEpoch = swProc.localEpochBase.add(new EpochOffset(0));
         if ( Epoch.Compare.gt(swc, procMinEpoch) ) {
            // we know swc <= base + MAX, since the sw value cannot "come from the future"
            assert EpochOffset.Compare.lte(newEpochOffset,
                                           swProc.epochOffsetManager.MAX_BASEOFFSET_VALUE());
            doUpdate = true;
         }
      }
      if ( doUpdate ) {
         // may have to grow line if we're adding a byteSized access to a compact line
         if ( byteSized && line.isCompact() ) {
            MetadataByteAddress swProcMba = new MetadataByteAddress(mba, swProc.id);
            CompactMetadataAddresses cma = getCompactAddresses(swProcMba);
            MemoryResponse<Line> resp = swProc.growLine(cma);
            // update lines map
            lines.put(swProc, resp.lineHit);
            line = resp.lineHit;
         }

         line.recordLocalAccess(mtype, mba, newEpochOffset, byteSized, toFree,
                                params.useTypeSafetyOpt(), false/* don't clobber reads */);

         if ( params.useClockGC() ) {
            for ( EpochOffset eo : toFree ) {
               assert eo.touched();
               swProc.epochOffsetManager.decrRefCount(eo);
            }
         }
         toFree.clear();
         return true;
      }

      return false;
   }

   /**
    * setup an in-hw-status of Everything for this access
    * 
    * @param clearSwMd
    *           whether to clear software metadata
    */
   private void setEverythingInHardwareNow(final MetadataAccess access, boolean clearSwMd) {
      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         UnownedMetadataByteAddress umba = a.addr().getUnowned();

         inHardwareStatus.put(umba, InHardwareStatus.EVERYTHING);

         if ( clearSwMd ) this.globalMetaDataTable.remove(umba);
      }
   }

   private class Verifier {

      private void verifyCacheInvariants() {
         verifyCacheIndexing();
         refCounts();
         cacheExclusivity();
         inHardwareStatus();
         compactMetadata();
         metadataOwnership();
      }

      /** Verify that only a single processor has valid metadata for these access bytes */
      private void singleWriterInvariant(final MetadataAccess access,
                                         Map<Processor<Line>, Line> lines) {
         for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
            Permissions.getHardwareReadWriteClocks(a.addr(), lines, allProcessors.length);
         }
      }

      private void verifyCacheIndexing() {
         for ( Processor<Line> p : allProcessors ) {
            p.L1cache.verifyIndices();

            if ( p.L2cache != null ) p.L2cache.verifyIndices();
         }

         if ( L3cache != null ) L3cache.verifyIndices();
      }

      /** Verify that the ref counts match up with the cache contents. */
      private void refCounts() {
         assert RadishSim.XASSERTS;
         if ( !params.useClockGC() ) return;

         for ( Processor<Line> p : allProcessors ) {
            final Processor<Line> proc = p;
            final Multiset<EpochOffset> refCounts = HashMultiset.create();

            final RadishLine.EpochOffsetVisitor countEpochOffsets = new RadishLine.EpochOffsetVisitor() {
               @Override
               public void visit(final MemoryAccessType _, final MetadataByteAddress __,
                                 final ByteMetadata md) {
                  if ( md.epochOffset.touched() ) {
                     if ( !params.useTypeSafetyOpt() || md.byteSized ) {
                        assert proc.epochOffsetManager.isAllocated(md.epochOffset);
                        refCounts.add(md.epochOffset);
                     }
                  }
               }
            };

            HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
               @Override
               public void visit(Line line) {
                  if ( !line.isMetadata() ) return;
                  MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
                  // only our metadata should be in our private caches
                  if ( !mla.owner().equals(proc.id) ) return;
                  line.visitLocalEpochOffsets(countEpochOffsets);
               }
            };

            // walk our private caches and count up all the epoch offsets
            proc.L1cache.visitAllLines(lv);
            if ( proc.L2cache != null ) {
               proc.L2cache.visitAllLines(lv);
            }
            if ( proc.L3cache != null ) {
               proc.L3cache.visitAllLines(lv);
            }

            assert refCounts.equals(proc.epochOffsetManager.getRefCounts());
         }
      }

      /**
       * Verify that there is only one copy of line, and all processors agree that copy is in the
       * LLC.
       */
      private void lineIsOnlyInLLC(Line line) {
         assert RadishSim.XASSERTS;
         assert line.isMetadata() && line.getState() == MESIState.MESI_MODIFIED;
         MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
         CpuId lineOwner = mla.owner();
         MetadataByteAddress mba = new MetadataByteAddress(mla);

         for ( Processor<Line> p : allProcessors ) {

            MemoryResponse<Line> resp = p.L1cache.ownedMdSearch(mba, false);

            if ( llc() == CacheLevel.L3 || lineOwner.equals(p.id) ) {
               // if there's a shared LLC, everyone should find the line in the LLC
               // otherwise, we are searching the owner's private hierarchy

               // line should be in the LLC
               assert resp.whereHit == llc();
               // we should find the copy owned by lineOwner
               assert resp.lineHit.isMetadata()
                      && resp.lineHit.getState() == MESIState.MESI_MODIFIED;
               assert ((MetadataLineAddress) resp.lineHit.lineAddress()).owner().equals(lineOwner);

            } else { // private LLCs and searching a non-owner's hierarchy
               // no one else should have the owner's copy
               assert resp.whereHit == CacheLevel.MEMORY;
            }

         }
      }

      /** Private caches should contain only their processor's metadata */
      private void metadataOwnership() {
         final Integer index[] = { 0 };
         HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
            @Override
            public void visit(Line line) {
               if ( !line.isMetadata() ) return;

               MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
               assert mla.owner().equals(allProcessors[index[0]].id);
            }
         };

         for ( int i = 0; i < allProcessors.length; i++ ) {
            Processor<Line> p = allProcessors[i];
            index[0] = i;

            p.L1cache.visitAllLines(lv);
            if ( p.L2cache != null ) {
               p.L2cache.visitAllLines(lv);
            }
         }
      }

      /** Verify that a cache line object is not resident in multiple caches simultaneously. */
      private void cacheExclusivity() {
         assert RadishSim.XASSERTS;

         LinkedList<HashSet<Line>> cacheContents = new LinkedList<HashSet<Line>>();
         final HashSet<Line> set = new HashSet<Line>();
         HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
            @Override
            public void visit(Line line) {
               set.add(line);
            }
         };

         for ( Processor<Line> p : allProcessors ) {
            p.L1cache.visitAllLines(lv);
            cacheContents.add(new HashSet<Line>(set));
            set.clear();

            if ( p.L2cache != null ) {
               p.L2cache.visitAllLines(lv);
               cacheContents.add(new HashSet<Line>(set));
               set.clear();
            }
         }

         if ( L3cache != null ) {
            L3cache.visitAllLines(lv);
            cacheContents.add(new HashSet<Line>(set));
            set.clear();
         }

         // every set in cacheContents should be disjoint from every other

         for ( HashSet<Line> a : cacheContents ) {
            for ( HashSet<Line> b : cacheContents ) {
               if ( a != b ) {
                  HashSet<Line> _a = new HashSet<Line>(a);
                  boolean changed = _a.retainAll(b); // a intersect b
                  assert changed;
                  assert _a.size() == 0;
               }
            }
         }

      }

      /** Verify that in-hw status accurately summarizes the current machine state */
      private void inHardwareStatus() {
         assert RadishSim.XASSERTS;

         final Map<UnownedMetadataLineAddress, List<Line>> lines = new HashMap<UnownedMetadataLineAddress, List<Line>>();

         HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
            @Override
            public void visit(Line line) {
               if ( !line.isMetadata() ) return;

               MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
               UnownedMetadataLineAddress umla = new UnownedMetadataLineAddress(mla.get());

               List<Line> list = lines.get(umla);
               if ( list == null ) {
                  list = new LinkedList<Line>();
                  lines.put(umla, list);
               }

               list.add(line);
            }
         };

         for ( Processor<Line> p : allProcessors ) {
            p.L1cache.visitAllLines(lv);
            if ( p.L2cache != null ) {
               p.L2cache.visitAllLines(lv);
            }
         }
         if ( L3cache != null ) {
            L3cache.visitAllLines(lv);
         }

         for ( UnownedMetadataLineAddress umla : lines.keySet() ) {
            List<Line> list = lines.get(umla);
            assert list != null;

            MetadataLineAddress mla = new MetadataLineAddress(umla.get(), Processor.this.id);
            // NB: use expanded sizing to be conservative
            MetadataAccess a = new MetadataAccess(
                                                  new MetadataByteAddress(mla),
                                                  MemorySystemConstants.DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.EXPANDED));
            for ( ; a.keepGoing(); a.incr() ) {

               // verify single-writer invariant
               int numWrites = 0, numLeafReads = 0;
               for ( Line l : list ) {
                  ByteMetadata md = l.getLocalEpochOffset(MemoryAccessType.WRITE, a.addr());
                  if ( md.epochOffset.touched() ) numWrites++;

                  md = l.getLocalEpochOffset(MemoryAccessType.READ, a.addr());
                  if ( md.epochOffset.touched() ) {
                     if ( md.leafRead ) {
                        numLeafReads++;
                     }
                  }
               }
               assert numWrites <= 1;

               InHardwareStatus ihs = inHardwareStatus.get(a.addr().getUnowned());

               if ( ihs == null ) {
                  // can have some readers in hw, nothing more
                  assert numWrites == 0;
                  continue;
               }

               /*
                * NB: on a sw check, we defoliate based on sw information, even if that sw info is
                * unrepresentable in hardware. Thus, a situation can arise in which we have only
                * non-leaf reads in hw, and the check below is too strong.
                */
               // if ( numReads > 0 ) assert numLeafReads > 0;

               switch ( ihs ) {
               case ALL_LEAF_READS: // there should never be any writes in hw
                  assert numWrites == 0;
                  break;
               case LAST_WRITE:
                  break;
               case EVERYTHING:
                  // could be hw write, or all-reads-in-hw, or an unaccessed location
                  break;
               }

            }

         }

      }

      /** Verify that md line addresses are consistent with their metadata type (compact/expanded). */
      private void compactMetadata() {

         HierarchicalCache.LineVisitor<Line> lv = new HierarchicalCache.LineVisitor<Line>() {
            @Override
            public void visit(Line line) {
               if ( !line.isMetadata() ) return;

               MetadataLineAddress mla = (MetadataLineAddress) line.lineAddress();
               if ( line.isCompact() ) {
                  assert MemorySystemConstants.COMPACT_MD_ALIGNED(mla.get());
               } else {
                  assert 0 == (mla.get() & MemorySystemConstants.LINE_OFFSET_MASK());
               }
            }
         };

         for ( Processor<Line> p : allProcessors ) {
            p.L1cache.visitAllLines(lv);
            if ( p.L2cache != null ) {
               p.L2cache.visitAllLines(lv);
            }
         }
         if ( L3cache != null ) {
            L3cache.visitAllLines(lv);
         }
      }

   } // end class Verifier

   private Verifier Verify = new Verifier();

} // end class Processor