package edu.washington.cs.radishsim;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import static edu.washington.cs.radishsim.MemoryAccessType.*;

public class SoftwareMetadata {

   /** Map byte addresses to their software VCs */
   private final Map<UnownedMetadataByteAddress, SoftwareVCPair> VcOfByte;

   /** The size of each VC (i.e. number of caches in the system) */
   private final int                                             VC_SIZE;

   /** The maximum number of out-of-cache metadata elements reached over the life of the program. */
   private long                                                  maxVcMapSize = 0;

   public SoftwareMetadata(final int vcsize) {
      VC_SIZE = vcsize;
      this.VcOfByte = new HashMap<UnownedMetadataByteAddress, SoftwareVCPair>();
   }

   /** Update the VC map with these new VC entries from hardware */
   public void update(final CpuId cpuid, final UnownedMetadataByteAddress byteAddr,
                      final MemoryAccessType mtype, final Epoch clock, final boolean byteSized) {
      assert !clock.equals(Epoch.EMPTY_EPOCH);

      SoftwareVCPair vcp = this.VcOfByte.get(byteAddr);

      if ( vcp == null ) { // create a new SoftwareVCPair
         vcp = new SoftwareVCPair(VC_SIZE);
         this.VcOfByte.put(byteAddr, vcp);
         this.maxVcMapSize = Math.max(this.maxVcMapSize, this.VcOfByte.size());
      }
      // vcp points to byteAddr's SoftwareVCPair now

      short i = cpuid.get();

      switch ( mtype ) {
      case READ:
         assert Epoch.Compare.gte(clock, vcp.read[i]);
         vcp.read[i] = clock;
         break;
      case WRITE:
         vcp.setWrite(cpuid, clock);
         break;
      }

      vcp.byteSized = byteSized;
   }

   /**
    * Simulates a FastTrack check on the given location, returning a list of DataAccesses that
    * represent the memory locations that FastTrack would touch.
    * 
    * @param svc
    *           the software metadata for the location being checked
    * @param umba
    *           the address of the location being checked
    * @param at
    *           whether the location is being read or written
    * @param me
    *           the id of the cpu performing the check
    * @param myVC
    *           the current per-core VC for processor me
    */
   public static List<DataAccess> doFastTrackCheck(SoftwareVCPair svc,
                                                   UnownedMetadataByteAddress umba,
                                                   MemoryAccessType at, CpuId me, Epoch[] myVC) {
      List<DataAccess> accesses = new LinkedList<DataAccess>();
      final long swMdAddr = MemorySystemConstants.SOFTWARE_MD_ADDR_OF_HW_MD(umba);
      final long lockAddr = swMdAddr;
      final long writeFTEpochAddr = lockAddr + 1;
      final long readFTEpochAddr = writeFTEpochAddr + 4;
      final long readVCAddr = readFTEpochAddr + 4;
      final Epoch myEpoch = myVC[me.get()];
      final long threadVCAddr = MemorySystemConstants.SOFTWARE_MD_VC(me.get());

      /* grab lock */
      accesses.add(new DataAccess(WRITE, new DataByteAddress(lockAddr), 1));

      if ( svc == null ) {
         // location hasn't been accessed yet
         accesses.add(new DataAccess(WRITE, new DataByteAddress(lockAddr), 1));
         
         // initialize VC
         accesses.add(new DataAccess(WRITE, new DataByteAddress(writeFTEpochAddr), 1));
         accesses.add(new DataAccess(WRITE, new DataByteAddress(readFTEpochAddr), 1));
         
         return accesses;
      }
      
      /*
       * NB: code behind // comments is actual FastTrack code.
       * 
       * Also, "FTEpoch" refers to the FastTrack notion of epoch (i.e., tid+clock value), not the
       * Radish notion (i.e., clock value).
       */

      int numConcurrentReads = 0;
      int readIndex = -1;
      for ( int i = 0; i < svc.read.length; i++ ) {
         if ( svc.read[i].touched() ) {
            numConcurrentReads++;
            readIndex = i;
         }
      }

      switch ( at ) {
      case WRITE:

         /* fast-path: we wrote last */
         accesses.add(new DataAccess(READ, new DataByteAddress(writeFTEpochAddr), 1));
         if ( svc.getWriter() == me && svc.getWrite().equals(myEpoch) ) {
            break;
         }
         //final int lastWriteEpoch = x.lastWrite;
         //if (lastWriteEpoch == tdEpoch) {
         //   return;
         //}

         /* check that we HA the last write */
         accesses.add(new DataAccess(READ, new DataByteAddress(threadVCAddr), 1));
         //final int lastWriter = Epoch.tid(lastWriteEpoch);           
         //if (lastWriteEpoch > tdCV.get(lastWriter)) {
         //   error(fae, 1, "write-by-thread-", lastWriter, "write-by-thread-", td.getTid());
         //}

         /* check for <= 1 last-reads */
         accesses.add(new DataAccess(READ, new DataByteAddress(readFTEpochAddr), 1));
         accesses.add(new DataAccess(READ, new DataByteAddress(threadVCAddr), 1));

         if ( numConcurrentReads <= 1 ) {} /* fast-path: 0 or 1 last-reads */
         else { /* slow path: numConcurrentReads > 1 */
            for ( int i = 0; i < svc.read.length; i++ ) {
               accesses.add(new DataAccess(READ, new DataByteAddress(readVCAddr + (4 * i)), 1));
               accesses.add(new DataAccess(READ, new DataByteAddress(threadVCAddr), 1));
            }
         }
         //final int lastReadEpoch = x.lastRead;           
         //if (lastReadEpoch != Epoch.EMPTY) {
         //  final int lastReader = Epoch.tid(lastReadEpoch);
         //  if (lastReader != td.getTid() && lastReadEpoch > tdCV.get(lastReader)) {
         //    error(fae, 2, "read-by-thread-", lastReader, "write-by-thread-", td.getTid());
         //  }
         //} else {
         //  if (x.anyGt(tdCV)) {
         //    for (int prevReader = x.nextGt(tdCV, 0); prevReader > -1; prevReader = x.nextGt(tdCV, prevReader + 1)) {
         //      if (prevReader != td.getTid()) {
         //        error(fae, 3, "read-by-thread-", prevReader, "write-by-thread-", td.getTid());
         //      }
         //    }
         //  }
         //}

         /* update read/write FT epochs */
         accesses.add(new DataAccess(WRITE, new DataByteAddress(writeFTEpochAddr), 1));
         //x.lastWrite = tdEpoch;
         accesses.add(new DataAccess(WRITE, new DataByteAddress(readFTEpochAddr), 1));
         //x.lastRead = tdEpoch;
         break;

      case READ:

         /* fast-path: we are the sole last-read */
         accesses.add(new DataAccess(READ, new DataByteAddress(readFTEpochAddr), 1));
         if ( !svc.hasWrite() && numConcurrentReads == 1 && svc.read[me.get()].equals(myEpoch) ) {
            break;
         } else if ( svc.read[me.get()].equals(myEpoch) ) { /* we're among the last-reads */
            break;
         }
         //final int lastReadEpoch = x.lastRead;
         //if (lastReadEpoch == tdEpoch) {
         //   return;
         //} else if (lastReadEpoch == Epoch.EMPTY) {
         //   if (x.get(td.getTid()) == tdEpoch) {
         //      return;
         //   }
         //}

         /* check that we HA the last write */
         accesses.add(new DataAccess(READ, new DataByteAddress(writeFTEpochAddr), 1));
         accesses.add(new DataAccess(READ, new DataByteAddress(threadVCAddr), 1));
         //final int lastWriteEpoch = x.lastWrite;
         //final int lastWriter = Epoch.tid(lastWriteEpoch);
         //if (lastWriteEpoch > tdCV.get(lastWriter)) {
         //   error(fae, 4, "write-by-thread-", lastWriter, "read-by-thread-", td.getTid());
         //}

         switch ( numConcurrentReads ) {
         case 0: /* just update read FT epoch */
            accesses.add(new DataAccess(WRITE, new DataByteAddress(readFTEpochAddr), 1));
            break;
         case 1: /* do we HA the last-read? */
            assert readIndex > -1;
            accesses.add(new DataAccess(READ, new DataByteAddress(threadVCAddr), 1));
            if ( Epoch.ha(myEpoch, myVC[readIndex]) ) {
               accesses.add(new DataAccess(WRITE, new DataByteAddress(readFTEpochAddr), 1));
            } else { // inflate read VC
               for ( int i = 0; i < svc.read.length; i++ ) {
                  accesses.add(new DataAccess(WRITE, new DataByteAddress(readVCAddr + (4 * i)), 1));
               }
               accesses.add(new DataAccess(WRITE, new DataByteAddress(readFTEpochAddr), 1));
            }
            break;
         default: /* just update our read VC entry */
            accesses.add(new DataAccess(WRITE, new DataByteAddress(readVCAddr), 1));
            break;
         }

         //if (lastReadEpoch != Epoch.EMPTY) {
         //   final int lastReader = Epoch.tid(lastReadEpoch);
         //   if (lastReadEpoch <= tdCV.get(lastReader)) {
         //      x.lastRead = tdEpoch;
         //   } else {
         //      x.makeCV(INIT_CV_SIZE);
         //      x.set(lastReader, lastReadEpoch);
         //      x.set(td.getTid(), tdEpoch);
         //      x.lastRead = Epoch.EMPTY;
         //   }
         //} else {
         //   x.set(td.getTid(), tdEpoch);
         //}
         break;

      default:
         assert false;
      }

      /* release lock */
      accesses.add(new DataAccess(WRITE, new DataByteAddress(lockAddr), 1));
      
      return accesses;
   }

   /**
    * @return the software read/write epochs for the given location. Returns null if there is no
    *         software metadata for this location.
    */
   public SoftwareVCPair get(final UnownedMetadataByteAddress umba) {
      return this.VcOfByte.get(umba);
   }

   /**
    * Remove the software read/write epochs for the given location.
    */
   public void remove(final UnownedMetadataByteAddress umba) {
      this.VcOfByte.remove(umba);
   }

   public void dump(Writer wr, String prefix, String suffix) throws IOException {
      wr.write(prefix + "'MaxLiveSoftwareVCs': " + this.maxVcMapSize + suffix);
   }

} // end class OutOfCacheMetadata
