package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.Random;

import org.junit.Before;
import org.junit.Test;

public class RemoteAccessesUpdateLRUTests {
   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_ASSOC             = 2;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 lines, 2 sets
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {
         
         @Override
         int numProcessors() {
            return CORES;
         }
         
         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }
         
         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }
         
         @Override
         boolean remoteAccessesAffectLRU() {
            return true;
         }
      };

      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B
   
   @Test
   public void test_remoteAccessUpdatesLRU() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      m.cacheRead(P0, 0, 1);
      // p0 set 0 is [0md=0, 0d]
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      
      m.cacheRead(P0, 4, 1);
      // p0 set 0 is [4md=8, 0md=0] (0d was evicted)
      // p0 set 1 is [4d, _]
      assertEquals(2, p0.stats.pc_SwVcChecksTotal.get());
      
      m.cacheRead(P1, 0, 1);
      // p0 set 0 is [0md=0, 4md=8]
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      
      m.cacheRead(P0, 12, 1);
      // p0 set 0 is [12md=24, 0md=0] (4md=8 was evicted)
      // p0 set 1 is [12d, 4d]
      MemoryResponse<RadishLine> res = p0.L1cache.ownedMdSearch(new MetadataByteAddress(0, P0), false);
      assertNotNull( res.lineHit );
      res = p0.L1cache.ownedMdSearch(new MetadataByteAddress(4, P0), false);
      assertNull( res.lineHit );
   }

   @Test
   public void test_randomAccesses() {
      Random rng = new Random(4321);

      for ( int i = 0; i < 50000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, read ? SynchronizationOperationType.SOURCE
                              : SynchronizationOperationType.SOURCE, false, addr);

         int size = rng.nextBoolean() ? 1 : 2;
         if ( read ) m.cacheRead(cpu, addr, size);
         else m.cacheWrite(cpu, addr, size);
      }
   }

}
